from __future__ import annotations

import os

from gpt4all import GPT4All
from .models import get_models
from ..typing import Messages

MODEL_LIST: dict[str, dict] = None

def find_model_dir(model_file: str) -> str:
    local_dir = os.path.dirname(os.path.abspath(__file__))
    project_dir = os.path.dirname(os.path.dirname(local_dir))

    new_model_dir = os.path.join(project_dir, "models")
    new_model_file = os.path.join(new_model_dir, model_file)
    if os.path.isfile(new_model_file):
        return new_model_dir

    old_model_dir = os.path.join(local_dir, "models")
    old_model_file = os.path.join(old_model_dir, model_file)
    if os.path.isfile(old_model_file):
        return old_model_dir

    working_dir = "./"
    for root, dirs, files in os.walk(working_dir):
        if model_file in files:
            return root

    return new_model_dir

class LocalProvider:
    @staticmethod
    def create_completion(model: str, messages: Messages, stream: bool = False, **kwargs):
        global MODEL_LIST
        if MODEL_LIST is None:
            MODEL_LIST = get_models()
        if model not in MODEL_LIST:
            raise ValueError(f'Model "{model}" not found / not yet implemented')

        model = MODEL_LIST[model]
        model_file = model["path"]
        model_dir = find_model_dir(model_file)
        if not os.path.isfile(os.path.join(model_dir, model_file)):
            print(f'Model file "models/{model_file}" not found.')
            download = input(f"Do you want to download {model_file}? [y/n]: ")
            if download in ["y", "Y"]:
                GPT4All.download_model(model_file, model_dir)
            else:
                raise ValueError(f'Model "{model_file}" not found.')

        model = GPT4All(model_name=model_file,
                        #n_threads=8,
                        verbose=False,
                        allow_download=False,
                        model_path=model_dir)

        system_message = "\n".join(message["content"] for message in messages if message["role"] == "system")
        if system_message:
            system_message = "A chat between a curious user and an artificial intelligence assistant."

        prompt_template = "USER: {0}\nASSISTANT: "
        conversation    = "\n" . join(
            f"{message['role'].upper()}: {message['content']}"
            for message in messages
            if message["role"] != "system"
        ) + "\nASSISTANT: "

        def should_not_stop(token_id: int, token: str):
            return "USER" not in token

        with model.chat_session(system_message, prompt_template):
            if stream:
                for token in model.generate(conversation, streaming=True, callback=should_not_stop):
                    yield token
            else:
                yield model.generate(conversation, callback=should_not_stop)


# from __future__ import annotations

# import os
# import re
# import json

# try:
#     from curl_cffi.requests import Session
#     has_curl_cffi = True
# except ImportError:
#     has_curl_cffi = False

# from ...typing import AsyncResult, Messages
# from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
# from ..helper import format_prompt
# from ...errors import MissingRequirementsError

# class ChatGptEs(AsyncGeneratorProvider, ProviderModelMixin):
#     url = "https://chatgpt.es"
#     api_endpoint = "https://chatgpt.es/wp-admin/admin-ajax.php"
    
#     working = False
#     supports_stream = True
#     supports_system_message = False
#     supports_message_history = False
    
#     default_model = 'gpt-4o'
#     models = ['gpt-4', default_model, 'gpt-4o-mini']
    
#     SYSTEM_PROMPT = "Your default language is English. Always respond in English unless the user's message is in a different language. If the user's message is not in English, respond in the language of the user's message. Maintain this language behavior throughout the conversation unless explicitly instructed otherwise. User input:"
    
#     @classmethod
#     async def create_async_generator(
#         cls,
#         model: str,
#         messages: Messages,
#         proxy: str = None,
#         **kwargs
#     ) -> AsyncResult:
#         if not has_curl_cffi:
#             raise MissingRequirementsError('Install or update "curl_cffi" package | pip install -U curl_cffi')
            
#         model = cls.get_model(model)
#         prompt = f"{cls.SYSTEM_PROMPT} {format_prompt(messages)}"
        
#         # Use curl_cffi with automatic Cloudflare bypass
#         session = Session()
#         session.headers.update({
#             "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36",
#             "referer": cls.url,
#             "origin": cls.url,
#             "accept": "*/*",
#             "accept-language": "en-US,en;q=0.9",
#             "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
#             "x-requested-with": "XMLHttpRequest",
#         })
        
#         if proxy:
#             session.proxies = {"https": proxy, "http": proxy}
        
#         # First request to get nonce and post_id
#         initial_response = session.get(cls.url, impersonate="chrome110")
#         initial_text = initial_response.text
        
#         # More comprehensive nonce extraction
#         nonce_patterns = [
#             r'<input\s+type=[\'"]hidden[\'"]\s+name=[\'"]_wpnonce[\'"]\s+value=[\'"]([^\'"]+)[\'"]',
#             r'"_wpnonce":"([^"]+)"',
#             r'var\s+wpaicg_nonce\s*=\s*[\'"]([^\'"]+)[\'"]',
#             r'wpaicg_nonce\s*:\s*[\'"]([^\'"]+)[\'"]'
#         ]
        
#         nonce_ = None
#         for pattern in nonce_patterns:
#             match = re.search(pattern, initial_text)
#             if match:
#                 nonce_ = match.group(1)
#                 break
                
#         if not nonce_:
#             # Try to find any nonce-like pattern as a last resort
#             general_nonce = re.search(r'nonce[\'"]?\s*[=:]\s*[\'"]([a-zA-Z0-9]+)[\'"]', initial_text)
#             if general_nonce:
#                 nonce_ = general_nonce.group(1)
#             else:
#                 # Fallback, but this likely won't work
#                 nonce_ = "8cf9917be2"
        
#         # Look for post_id in HTML
#         post_id_patterns = [
#             r'<input\s+type=[\'"]hidden[\'"]\s+name=[\'"]post_id[\'"]\s+value=[\'"]([^\'"]+)[\'"]',
#             r'"post_id":"([^"]+)"',
#             r'var\s+post_id\s*=\s*[\'"]?(\d+)[\'"]?'
#         ]
        
#         post_id = None
#         for pattern in post_id_patterns:
#             match = re.search(pattern, initial_text)
#             if match:
#                 post_id = match.group(1)
#                 break
                
#         if not post_id:
#             post_id = "106"  # Default from curl example
        
#         client_id = os.urandom(5).hex()
        
#         # Prepare data
#         data = {
#             '_wpnonce': nonce_,
#             'post_id': post_id,
#             'url': cls.url,
#             'action': 'wpaicg_chat_shortcode_message',
#             'message': prompt,
#             'bot_id': '0',
#             'chatbot_identity': 'shortcode',
#             'wpaicg_chat_client_id': client_id,
#             'wpaicg_chat_history': json.dumps([f"Human: {prompt}"])
#         }

#         # Execute POST request
#         response = session.post(
#             cls.api_endpoint,
#             data=data,
#             impersonate="chrome110"
#         )
        
#         if response.status_code != 200:
#             raise ValueError(f"Error: {response.status_code} - {response.text}")
        
#         result = response.json()
#         if "data" in result:
#             if isinstance(result['data'], str) and "Du musst das Kästchen anklicken!" in result['data']:
#                 raise ValueError(result['data'])
#             yield result['data']
#         else:
#             raise ValueError(f"Unexpected response format: {result}")


from __future__ import annotations

import os
import re
import json
import aiohttp

from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import format_prompt
from ...errors import MissingRequirementsError

class ChatGptEs(AsyncGeneratorProvider, ProviderModelMixin):
    url = "https://chatgpt.es"
    api_endpoint = "https://chatgpt.es/wp-admin/admin-ajax.php"
    
    working = False
    supports_stream = True
    supports_system_message = False
    supports_message_history = False
    
    default_model = 'gpt-4o'
    models = ['gpt-4', default_model, 'gpt-4o-mini']
    
    SYSTEM_PROMPT = "Your default language is English. Always respond in English unless the user's message is in a different language. If the user's message is not in English, respond in the language of the user's message. Maintain this language behavior throughout the conversation unless explicitly instructed otherwise. User input:"
    
    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        prompt = f"{cls.SYSTEM_PROMPT} {format_prompt(messages)}"
        
        # Prepare headers
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/140.0.0.0 Safari/537.36",
            "referer": cls.url,
            "origin": cls.url,
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
        }
        
        async with aiohttp.ClientSession(headers=headers) as session:
            # Step 1: Fetch nonce and post_id
            async with session.get(cls.url) as initial_response:
                if initial_response.status != 200:
                    raise ValueError(f"Failed to fetch initial page: {initial_response.status}")
                initial_text = await initial_response.text()
            
            # Extract nonce
            nonce_patterns = [
                r'<input\s+type=[\'"]hidden[\'"]\s+name=[\'"]_wpnonce[\'"]\s+value=[\'"]([^\'"]+)[\'"]',
                r'"_wpnonce":"([^"]+)"',
                r'var\s+wpaicg_nonce\s*=\s*[\'"]([^\'"]+)[\'"]',
                r'wpaicg_nonce\s*:\s*[\'"]([^\'"]+)[\'"]'
            ]
            nonce_ = None
            for pattern in nonce_patterns:
                match = re.search(pattern, initial_text)
                if match:
                    nonce_ = match.group(1)
                    break
            if not nonce_:
                raise ValueError("Failed to extract _wpnonce from the initial page.")
            
            # Extract post_id
            post_id_patterns = [
                r'<input\s+type=[\'"]hidden[\'"]\s+name=[\'"]post_id[\'"]\s+value=[\'"]([^\'"]+)[\'"]',
                r'"post_id":"([^"]+)"',
                r'var\s+post_id\s*=\s*[\'"]?(\d+)[\'"]?'
            ]
            post_id = None
            for pattern in post_id_patterns:
                match = re.search(pattern, initial_text)
                if match:
                    post_id = match.group(1)
                    break
            if not post_id:
                post_id = "106"  # Default from curl example
            
            client_id = os.urandom(5).hex()
            
            # Step 2: Send POST request
            data = aiohttp.FormData()
            data.add_field('_wpnonce', nonce_)
            data.add_field('post_id', post_id)
            data.add_field('url', cls.url)
            data.add_field('action', 'wpaicg_chat_shortcode_message')
            data.add_field('message', prompt)
            data.add_field('bot_id', '0')
            data.add_field('chatbot_identity', 'shortcode')
            data.add_field('wpaicg_chat_client_id', client_id)
            data.add_field('wpaicg_chat_history', json.dumps([f"Human: {prompt}"]))

            async with session.post(cls.api_endpoint, data=data) as response:
                if response.status != 200:
                    error_text = await response.text()
                    raise ValueError(f"Error: {response.status} - {error_text}")
                
                result = await response.json()
                if "data" in result:
                    if isinstance(result['data'], str) and "Du musst das Kästchen anklicken!" in result['data']:
                        raise ValueError(result['data'])
                    yield result['data']
                else:
                    raise ValueError(f"Unexpected response format: {result}")


from __future__ import annotations

import json

from ..typing import AsyncResult, Messages, MediaListType
from ..client.service import get_model_and_provider
from ..client.helper import filter_json
from .base_provider import AsyncGeneratorProvider
from .response import ToolCalls, FinishReason, Usage

class ToolSupportProvider(AsyncGeneratorProvider):
    working = True

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        stream: bool = True,
        media: MediaListType = None,
        tools: list[str] = None,
        response_format: dict = None,
        **kwargs
    ) -> AsyncResult:
        provider = None
        if ":" in model:
            provider, model = model.split(":", 1)
        model, provider = get_model_and_provider(
            model, provider,
            stream, logging=False,
            has_images=media is not None
        )
        if tools is not None:
            if len(tools) > 1:
                raise ValueError("Only one tool is supported.")
            if response_format is None:
                response_format = {"type": "json"}
            tools = tools.pop()
            lines = ["Respone in JSON format."]
            properties = tools["function"]["parameters"]["properties"]
            properties = {key: value["type"] for key, value in properties.items()}
            lines.append(f"Response format: {json.dumps(properties, indent=2)}")
            messages = [{"role": "user", "content": "\n".join(lines)}] + messages

        finish = None
        chunks = []
        has_usage = False
        async for chunk in provider.async_create_function(
            model,
            messages,
            stream=stream,
            media=media,
            response_format=response_format,
            **kwargs
        ):
            if isinstance(chunk, str):
                chunks.append(chunk)
            elif isinstance(chunk, Usage):
                yield chunk
                has_usage = True
            elif isinstance(chunk, FinishReason):
                finish = chunk
                break
            else:
                yield chunk

        if not has_usage:
            yield Usage(completion_tokens=len(chunks), total_tokens=len(chunks))

        chunks = "".join(chunks)
        if tools is not None:
            yield ToolCalls([{
                "id": "",
                "type": "function",
                "function": {
                    "name": tools["function"]["name"],
                    "arguments": filter_json(chunks)
                }
            }])
        yield chunks

        if finish is not None:
            yield finish


from __future__ import annotations

import json
from aiohttp import ClientSession

from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import get_random_string, format_prompt

class AiChatOnline(AsyncGeneratorProvider, ProviderModelMixin):
    site_url = "https://aichatonline.org"
    url = "https://aichatonlineorg.erweima.ai"
    api_endpoint = "/aichatonline/api/chat/gpt"
    working = False
    default_model = 'gpt-4o-mini'

    @classmethod
    async def grab_token(
        cls,
        session: ClientSession,
        proxy: str
    ):
        async with session.get(f'https://aichatonlineorg.erweima.ai/api/v1/user/getUniqueId?canvas=-{get_random_string()}', proxy=proxy) as response:
            response.raise_for_status()
            return (await response.json())['data']
        
    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        **kwargs
    ) -> AsyncResult:
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/119.0",
            "Accept-Language": "de,en-US;q=0.7,en;q=0.3",
            "Accept-Encoding": "gzip, deflate, br",
            "Referer": f"{cls.url}/chatgpt/chat/",
            "Content-Type": "application/json",
            "Origin": cls.url,
            "Alt-Used": "aichatonline.org",
            "Connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "TE": "trailers"
        }
        async with ClientSession(headers=headers) as session:
            data = {
                "conversationId": get_random_string(),
                "prompt": format_prompt(messages),
            }
            headers['UniqueId'] = await cls.grab_token(session, proxy)
            async with session.post(f"{cls.url}{cls.api_endpoint}", headers=headers, json=data, proxy=proxy) as response:
                response.raise_for_status()
                async for chunk in response.content:
                    try:
                        yield json.loads(chunk)['data']['message']
                    except:
                        continue




from __future__ import annotations

from urllib.parse import quote
from aiohttp import ClientSession

from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import format_media_prompt, get_system_prompt
from ...image.copy_images import save_response_media
from ...providers.response import AudioResponse
from ...requests.raise_for_status import raise_for_status
from ...requests.aiohttp import get_connector
from ...requests import DEFAULT_HEADERS

class OpenAIFM(AsyncGeneratorProvider, ProviderModelMixin):
    label = "OpenAI.fm"
    url = "https://www.openai.fm"
    api_endpoint = "https://www.openai.fm/api/generate"
    working = True

    default_model = 'coral'
    voices = ['alloy', 'ash', 'ballad', 'coral', 'echo', 'fable', 'onyx', 'nova', 'sage', 'shimmer', 'verse']
    styles = ['friendly', 'patient_teacher', 'noir_detective', 'cowboy', 'calm', 'scientific_style']
    audio_models = {"gpt-4o-mini-tts": voices}
    model_aliases = {"gpt-4o-mini-tts": default_model}
    models = styles + voices

    @classmethod
    def get_grouped_models(cls):
        return [
            {"group":"Styles", "models": cls.styles},
            {"group":"Voices", "models": cls.voices},
        ]

    friendly = """Affect/personality: A cheerful guide 

Tone: Friendly, clear, and reassuring, creating a calm atmosphere and making the listener feel confident and comfortable.

Pronunciation: Clear, articulate, and steady, ensuring each instruction is easily understood while maintaining a natural, conversational flow.

Pause: Brief, purposeful pauses after key instructions (e.g., "cross the street" and "turn right") to allow time for the listener to process the information and follow along.

Emotion: Warm and supportive, conveying empathy and care, ensuring the listener feels guided and safe throughout the journey."""

    patient_teacher = """Accent/Affect: Warm, refined, and gently instructive, reminiscent of a friendly art instructor.

Tone: Calm, encouraging, and articulate, clearly describing each step with patience.

Pacing: Slow and deliberate, pausing often to allow the listener to follow instructions comfortably.

Emotion: Cheerful, supportive, and pleasantly enthusiastic; convey genuine enjoyment and appreciation of art.

Pronunciation: Clearly articulate artistic terminology (e.g., "brushstrokes," "landscape," "palette") with gentle emphasis.

Personality Affect: Friendly and approachable with a hint of sophistication; speak confidently and reassuringly, guiding users through each painting step patiently and warmly."""

    noir_detective = """Affect: a mysterious noir detective

Tone: Cool, detached, but subtly reassuring—like they've seen it all and know how to handle a missing package like it's just another case.

Delivery: Slow and deliberate, with dramatic pauses to build suspense, as if every detail matters in this investigation.

Emotion: A mix of world-weariness and quiet determination, with just a hint of dry humor to keep things from getting too grim.

Punctuation: Short, punchy sentences with ellipses and dashes to create rhythm and tension, mimicking the inner monologue of a detective piecing together clues."""

    cowboy = """Voice: Warm, relaxed, and friendly, with a steady cowboy drawl that feels approachable.

Punctuation: Light and natural, with gentle pauses that create a conversational rhythm without feeling rushed.

Delivery: Smooth and easygoing, with a laid-back pace that reassures the listener while keeping things clear.

Phrasing: Simple, direct, and folksy, using casual, familiar language to make technical support feel more personable.

Tone: Lighthearted and welcoming, with a calm confidence that puts the caller at ease."""

    calm = """Voice Affect: Calm, composed, and reassuring; project quiet authority and confidence.

Tone: Sincere, empathetic, and gently authoritative—express genuine apology while conveying competence.

Pacing: Steady and moderate; unhurried enough to communicate care, yet efficient enough to demonstrate professionalism.

Emotion: Genuine empathy and understanding; speak with warmth, especially during apologies ("I'm very sorry for any disruption...").

Pronunciation: Clear and precise, emphasizing key reassurances ("smoothly," "quickly," "promptly") to reinforce confidence.

Pauses: Brief pauses after offering assistance or requesting details, highlighting willingness to listen and support."""

    scientific_style = """Voice: Authoritative and precise, with a measured, academic tone.

Tone: Formal and analytical, maintaining objectivity while conveying complex information.

Pacing: Moderate and deliberate, allowing time for complex concepts to be processed.

Pronunciation: Precise articulation of technical terms and scientific vocabulary.

Pauses: Strategic pauses after introducing new concepts to allow for comprehension.

Emotion: Restrained enthusiasm for discoveries and findings, conveying intellectual curiosity."""

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        prompt: str = None,
        audio: dict = {},
        download_media: bool = True,
        **kwargs
    ) -> AsyncResult:
        default_instructions = get_system_prompt(messages)
        if model and hasattr(cls, model):
            default_instructions = getattr(cls, model)
            model = ""
        model = cls.get_model(model)
        voice = audio.get("voice", kwargs.get("voice", model))
        instructions = audio.get("instructions", kwargs.get("instructions", default_instructions))
        headers = {
            **DEFAULT_HEADERS,
            "referer": f"{cls.url}/"
        }
        prompt = format_media_prompt(messages, prompt)
        params = {
            "input": prompt,
            "prompt": instructions,
            "voice": voice
        }
        if not download_media:
            query = "&".join(f"{k}={quote(str(v))}" for k, v in params.items() if v is not None)
            yield AudioResponse(f"{cls.api_endpoint}?{query}")
            return
        async with ClientSession(headers=headers, connector=get_connector(proxy=proxy)) as session:            
            async with session.get(
                cls.api_endpoint,
                params=params
            ) as response:
                await raise_for_status(response)                
                async for chunk in save_response_media(response, prompt, [model, voice]):
                    yield chunk



from __future__ import annotations

import os
import asyncio
from typing import Any

try:
    from ...integration.markitdown import MarkItDown as MaItDo, StreamInfo
    has_markitdown = True
except ImportError:
    has_markitdown = False

from ...typing import AsyncResult, Messages, MediaListType
from ...tools.files import get_tempfile
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin

class MarkItDown(AsyncGeneratorProvider, ProviderModelMixin):
    working = has_markitdown

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        media: MediaListType = None,
        llm_client: Any = None,
        **kwargs
    ) -> AsyncResult:
        if media is None:
            raise ValueError("MarkItDown requires media to be provided.")
        if not has_markitdown:
            raise ImportError("MarkItDown is not installed. Please install it with `pip install markitdown`.")
        md = MaItDo()
        for file, filename in media:
            text = None
            try:
                result = md.convert(
                    file,
                    stream_info=StreamInfo(filename=filename) if filename else None,
                    llm_client=llm_client,
                    llm_model=model
                )
                if asyncio.iscoroutine(result.text_content):
                    text = await result.text_content
                else:
                    text = result.text_content
            except TypeError:
                copyfile = get_tempfile(file, filename)
                try:
                    result = md.convert(
                        copyfile, 
                        llm_client=llm_client,
                        llm_model=model
                    )
                    if asyncio.iscoroutine(result.text_content):
                        text = await result.text_content
                    else:
                        text = result.text_content
                finally:
                    os.remove(copyfile)
            text = text.split("### Audio Transcript:\n")[-1]
            if text:
                yield text


from __future__ import annotations

import json
import time
import random
import re
import os
import base64
import asyncio
from aiohttp import ClientSession, FormData

from ..typing import AsyncResult, Messages, MediaListType
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin
from .helper import format_prompt
from ..providers.response import JsonConversation, FinishReason, ImageResponse
from ..image import to_data_uri, is_data_an_media
from ..tools.media import merge_media


class Conversation(JsonConversation):
    """Manages all session-specific state for Opera Aria."""
    access_token: str = None
    refresh_token: str = None
    encryption_key: str = None
    expires_at: float = 0
    conversation_id: str = None
    is_first_request: bool = True
    
    def __init__(self, refresh_token: str = None):
        """Initializes a new session, generating a unique encryption key."""
        self.refresh_token = refresh_token
        self.encryption_key = self._generate_encryption_key()
        self.is_first_request = True
    
    def is_token_expired(self) -> bool:
        """Check if the current token has expired"""
        return time.time() >= self.expires_at
    
    def update_token(self, access_token: str, expires_in: int):
        """Update the access token and expiration time"""
        self.access_token = access_token
        self.expires_at = time.time() + expires_in - 60
    
    @staticmethod
    def _generate_encryption_key() -> str:
        """Generates a 32-byte, Base64-encoded key for the session."""
        random_bytes = os.urandom(32)
        return base64.b64encode(random_bytes).decode('utf-8')

    @staticmethod
    def generate_conversation_id() -> str:
        """Generate conversation ID in Opera Aria format"""
        parts = [
            ''.join(random.choices('0123456789abcdef', k=8)),
            ''.join(random.choices('0123456789abcdef', k=4)),
            '11f0',
            ''.join(random.choices('0123456789abcdef', k=4)),
            ''.join(random.choices('0123456789abcdef', k=12))
        ]
        return '-'.join(parts)


class OperaAria(AsyncGeneratorProvider, ProviderModelMixin):
    label = "Opera Aria"
    url = "https://play.google.com/store/apps/details?id=com.opera.browser"
    api_endpoint = "https://composer.opera-api.com/api/v1/a-chat"
    token_endpoint = "https://oauth2.opera-api.com/oauth2/v1/token/"
    signup_endpoint = "https://auth.opera.com/account/v2/external/anonymous/signup"
    upload_endpoint = "https://composer.opera-api.com/api/v1/images/upload"
    check_status_endpoint = "https://composer.opera-api.com/api/v1/images/check-status/"
    
    working = True
    needs_auth = False
    supports_stream = True
    supports_system_message = True
    supports_message_history = True
    
    default_model = 'aria'
    default_image_model = 'aria'
    image_models = ['aria']
    default_vision_model = 'aria'
    vision_models = ['aria']
    models = ['aria']
        
    @classmethod
    async def _generate_refresh_token(cls, session: ClientSession) -> str:
        headers = {
            "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "client_id": "ofa-client",
            "client_secret": "N9OscfA3KxlJASuIe29PGZ5RpWaMTBoy",
            "grant_type": "client_credentials",
            "scope": "anonymous_account"
        }
        async with session.post(cls.token_endpoint, headers=headers, data=data) as response:
            response.raise_for_status()
            anonymous_token_data = await response.json()
            anonymous_access_token = anonymous_token_data["access_token"]

        headers = {
            "User-Agent": "Mozilla 5.0 (Linux; Android 14) com.opera.browser OPR/89.5.4705.84314",
            "Authorization": f"Bearer {anonymous_access_token}",
            "Accept": "application/json",
            "Content-Type": "application/json; charset=utf-8",
        }
        data = {"client_id": "ofa", "service": "aria"}
        async with session.post(cls.signup_endpoint, headers=headers, json=data) as response:
            response.raise_for_status()
            signup_data = await response.json()
            auth_token = signup_data["token"]

        headers = {
            "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "auth_token": auth_token,
            "client_id": "ofa",
            "device_name": "GPT4FREE",
            "grant_type": "auth_token",
            "scope": "ALL"
        }
        async with session.post(cls.token_endpoint, headers=headers, data=data) as response:
            response.raise_for_status()
            final_token_data = await response.json()
            return final_token_data["refresh_token"]

    @classmethod
    def get_model(cls, model: str) -> str:
        return cls.model_aliases.get(model, cls.default_model)

    @classmethod
    async def get_access_token(cls, session: ClientSession, conversation: Conversation) -> str:
        if not conversation.refresh_token:
            conversation.refresh_token = await cls._generate_refresh_token(session)

        if conversation.access_token and not conversation.is_token_expired():
            return conversation.access_token
        
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0"
        }
        data = {
            "client_id": "ofa",
            "grant_type": "refresh_token",
            "refresh_token": conversation.refresh_token,
            "scope": "shodan:aria user:read"
        }
        async with session.post(cls.token_endpoint, headers=headers, data=data) as response:
            response.raise_for_status()
            result = await response.json()
            conversation.update_token(
                access_token=result["access_token"],
                expires_in=result.get("expires_in", 3600)
            )
            return result["access_token"]

    @classmethod
    async def check_upload_status(cls, session: ClientSession, access_token: str, image_id: str, max_attempts: int = 30):
        headers = {
            "Authorization": f"Bearer {access_token}",
            "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0",
        }
        url = f"{cls.check_status_endpoint}{image_id}"
        for _ in range(max_attempts):
            async with session.get(url, headers=headers) as response:
                response.raise_for_status()
                result = await response.json()
                if result.get("status") == "ok":
                    return
                if result.get("status") == "failed":
                    raise Exception(f"Image upload failed for {image_id}")
                await asyncio.sleep(0.5)
        raise Exception(f"Timeout waiting for image upload status for {image_id}")

    @classmethod
    async def upload_media(cls, session: ClientSession, access_token: str, media_data: bytes, filename: str) -> str:
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Origin": "opera-aria://ui",
            "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0",
        }
        form_data = FormData()
        if not filename:
            filename = str(int(time.time() * 1000))
        content_type = is_data_an_media(media_data, filename) or "application/octet-stream"
        form_data.add_field('image_file', media_data, filename=filename, content_type=content_type)
        
        async with session.post(cls.upload_endpoint, headers=headers, data=form_data) as response:
            response.raise_for_status()
            result = await response.json()
            image_id = result.get("image_id")
            if not image_id:
                raise Exception("No image_id returned from upload")
            await cls.check_upload_status(session, access_token, image_id)
            return image_id

    @classmethod
    def extract_image_urls(cls, text: str) -> list[str]:
        pattern = r'!\[\]\((https?://[^\)]+)\)'
        urls = re.findall(pattern, text)
        return [url.replace(r'\/', '/') for url in urls]

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        refresh_token: str = None,
        conversation: Conversation = None,
        return_conversation: bool = False,
        stream: bool = True,
        media: MediaListType = None,
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        
        if conversation is None:
            conversation = Conversation(refresh_token)
        elif refresh_token and not conversation.refresh_token:
            conversation.refresh_token = refresh_token
        
        async with ClientSession() as session:
            access_token = await cls.get_access_token(session, conversation)
            
            media_attachments = []
            merged_media = list(merge_media(media, messages))
            if merged_media:
                for media_data, media_name in merged_media:
                    try:
                        if isinstance(media_data, str) and media_data.startswith("data:"):
                            data_part = media_data.split(",", 1)[1]
                            media_bytes = base64.b64decode(data_part)
                        elif hasattr(media_data, 'read'):
                            media_bytes = media_data.read()
                        elif isinstance(media_data, (str, os.PathLike)):
                            with open(media_data, 'rb') as f:
                                media_bytes = f.read()
                        else:
                            media_bytes = media_data
                        
                        image_id = await cls.upload_media(session, access_token, media_bytes, media_name)
                        media_attachments.append(image_id)
                    except Exception:
                        continue
            
            headers = {
                "Accept": "text/event-stream" if stream else "application/json",
                "Authorization": f"Bearer {access_token}",
                "Content-Type": "application/json",
                "Origin": "opera-aria://ui",
                "User-Agent": "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36 OPR/89.0.0.0",
                "X-Opera-Timezone": "+03:00",
                "X-Opera-UI-Language": "en"
            }
            
            data = {
                "query": format_prompt(messages), "stream": stream, "linkify": True,
                "linkify_version": 3, "sia": True, "media_attachments": media_attachments,
                "encryption": {"key": conversation.encryption_key}
            }
            
            if not conversation.is_first_request and conversation.conversation_id:
                data["conversation_id"] = conversation.conversation_id
            
            async with session.post(cls.api_endpoint, headers=headers, json=data, proxy=proxy) as response:
                response.raise_for_status()
                
                if stream:
                    text_buffer, image_urls, finish_reason = [], [], None
                    
                    async for line in response.content:
                        if not line: continue
                        decoded = line.decode('utf-8').strip()
                        if not decoded.startswith('data: '): continue
                        
                        content = decoded[6:]
                        if content == '[DONE]': break
                        
                        try:
                            json_data = json.loads(content)
                            if 'message' in json_data:
                                message_chunk = json_data['message']
                                found_urls = cls.extract_image_urls(message_chunk)
                                if found_urls:
                                    image_urls.extend(found_urls)
                                else:
                                    text_buffer.append(message_chunk)
                            
                            if 'conversation_id' in json_data and json_data['conversation_id']:
                                conversation.conversation_id = json_data['conversation_id']
                            
                            if 'finish_reason' in json_data and json_data.get('finish_reason'):
                                finish_reason = json_data['finish_reason']

                        except json.JSONDecodeError:
                            continue
                    
                    if image_urls:
                        yield ImageResponse(image_urls, format_prompt(messages))
                    elif text_buffer:
                        yield "".join(text_buffer)
                    
                    if finish_reason:
                        yield FinishReason(finish_reason)

                else: # Non-streaming
                    json_data = await response.json()
                    if 'message' in json_data:
                        message = json_data['message']
                        image_urls = cls.extract_image_urls(message)
                        if image_urls:
                            yield ImageResponse(image_urls, format_prompt(messages))
                        else:
                            yield message
                    
                    if 'conversation_id' in json_data and json_data['conversation_id']:
                        conversation.conversation_id = json_data['conversation_id']
                    
                    if 'finish_reason' in json_data and json_data['finish_reason']:
                        yield FinishReason(json_data['finish_reason'])
                
                conversation.is_first_request = False
                
                if return_conversation:
                    yield conversation



from __future__ import annotations

import os
import sys
from pathlib import Path
from functools import lru_cache

@lru_cache(maxsize=1)
def get_config_dir() -> Path:
    """Get platform-appropriate config directory."""
    if sys.platform == "win32":
        return Path(os.environ.get("APPDATA", Path.home() / "AppData" / "Roaming"))
    elif sys.platform == "darwin":
        return Path.home() / "Library" / "Application Support"
    return Path.home() / ".config"

DEFAULT_PORT = 1337
DEFAULT_TIMEOUT = 600
DEFAULT_STREAM_TIMEOUT = 15

PACKAGE_NAME = "g4f"
CONFIG_DIR = get_config_dir() / PACKAGE_NAME
COOKIES_DIR = CONFIG_DIR / "cookies"
CUSTOM_COOKIES_DIR = "./har_and_cookies"
ORGANIZATION = "gpt4free"
GITHUB_REPOSITORY = f"xtekky/{ORGANIZATION}"
STATIC_DOMAIN = f"{PACKAGE_NAME}.dev"
STATIC_URL = f"https://{STATIC_DOMAIN}/"
DIST_DIR = f"./{STATIC_DOMAIN}/dist"
DEFAULT_MODEL = "openai/gpt-oss-120b"
JSDELIVR_URL = "https://cdn.jsdelivr.net/"
DOWNLOAD_URL = f"{JSDELIVR_URL}gh/{ORGANIZATION}/{STATIC_DOMAIN}/"
GITHUB_URL = f"https://raw.githubusercontent.com/{ORGANIZATION}/{STATIC_DOMAIN}/refs/heads/main/"


from __future__ import annotations

import uuid
import json
from aiohttp import ClientSession, BaseConnector

from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import get_connector
from ...requests import raise_for_status
from ...errors import RateLimitError

models = {
    "claude-3-5-sonnet-20241022": {
        "id": "claude-3-5-sonnet-20241022",
        "name": "claude-3-5-sonnet-20241022",
        "model": "claude-3-5-sonnet-20241022",
        "provider": "Anthropic",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 25.366666666666667,
    },
    "claude-3-5-sonnet-20241022-t": {
        "id": "claude-3-5-sonnet-20241022-t",
        "name": "claude-3-5-sonnet-20241022-t",
        "model": "claude-3-5-sonnet-20241022-t",
        "provider": "Anthropic",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 39.820754716981135,
    },
    "claude-3-7-sonnet-20250219": {
        "id": "claude-3-7-sonnet-20250219",
        "name": "claude-3-7-sonnet-20250219",
        "model": "claude-3-7-sonnet-20250219",
        "provider": "Anthropic",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 47.02970297029703,
    },
    "claude-3-7-sonnet-20250219-t": {
        "id": "claude-3-7-sonnet-20250219-t",
        "name": "claude-3-7-sonnet-20250219-t",
        "model": "claude-3-7-sonnet-20250219-t",
        "provider": "Anthropic",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 39.04289693593315,
    },
    "deepseek-v3": {
        "id": "deepseek-v3",
        "name": "deepseek-v3",
        "model": "deepseek-v3",
        "provider": "DeepSeek",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 40.484657419083646,
    },
    "gemini-1.0-pro-latest-123": {
        "id": "gemini-1.0-pro-latest-123",
        "name": "gemini-1.0-pro-latest-123",
        "model": "gemini-1.0-pro-latest-123",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 10,
    },
    "gemini-2.0-flash": {
        "id": "gemini-2.0-flash",
        "name": "gemini-2.0-flash",
        "model": "gemini-2.0-flash",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 216.44162436548223,
    },
    "gemini-2.0-flash-exp": {
        "id": "gemini-2.0-flash-exp",
        "name": "gemini-2.0-flash-exp",
        "model": "gemini-2.0-flash-exp",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 0,
        "tps": 0,
    },
    "gemini-2.0-flash-thinking-exp": {
        "id": "gemini-2.0-flash-thinking-exp",
        "name": "gemini-2.0-flash-thinking-exp",
        "model": "gemini-2.0-flash-thinking-exp",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 0,
        "tps": 0,
    },
    "gemini-2.5-flash-preview-04-17": {
        "id": "gemini-2.5-flash-preview-04-17",
        "name": "gemini-2.5-flash-preview-04-17",
        "model": "gemini-2.5-flash-preview-04-17",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 189.84010840108402,
    },
    "gemini-2.5-pro-official": {
        "id": "gemini-2.5-pro-official",
        "name": "gemini-2.5-pro-official",
        "model": "gemini-2.5-pro-official",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 91.00613496932516,
    },
    "gemini-2.5-pro-preview-03-25": {
        "id": "gemini-2.5-pro-preview-03-25",
        "name": "gemini-2.5-pro-preview-03-25",
        "model": "gemini-2.5-pro-preview-03-25",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 99.05660377358491,
        "tps": 45.050511247443765,
    },
    "gemini-2.5-pro-preview-05-06": {
        "id": "gemini-2.5-pro-preview-05-06",
        "name": "gemini-2.5-pro-preview-05-06",
        "model": "gemini-2.5-pro-preview-05-06",
        "provider": "Google",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 99.29617834394904,
    },
    "gpt-4-turbo-2024-04-09": {
        "id": "gpt-4-turbo-2024-04-09",
        "name": "gpt-4-turbo-2024-04-09",
        "model": "gpt-4-turbo-2024-04-09",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 1,
    },
    "gpt-4.1": {
        "id": "gpt-4.1",
        "name": "gpt-4.1",
        "model": "gpt-4.1",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 42.857142857142854,
        "tps": 19.58032786885246,
    },
    "gpt-4.1-mini": {
        "id": "gpt-4.1-mini",
        "name": "gpt-4.1-mini",
        "model": "gpt-4.1-mini",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 68.75,
        "tps": 12.677576601671309,
    },
    "gpt-4.1-mini-2025-04-14": {
        "id": "gpt-4.1-mini-2025-04-14",
        "name": "gpt-4.1-mini-2025-04-14",
        "model": "gpt-4.1-mini-2025-04-14",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 94.23076923076923,
        "tps": 8.297687861271676,
    },
    "gpt-4o-2024-11-20": {
        "id": "gpt-4o-2024-11-20",
        "name": "gpt-4o-2024-11-20",
        "model": "gpt-4o-2024-11-20",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 73.3955223880597,
    },
    "gpt-4o-mini-2024-07-18": {
        "id": "gpt-4o-mini-2024-07-18",
        "name": "gpt-4o-mini-2024-07-18",
        "model": "gpt-4o-mini-2024-07-18",
        "provider": "OpenAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 26.874455100261553,
    },
    "grok-3": {
        "id": "grok-3",
        "name": "grok-3",
        "model": "grok-3",
        "provider": "xAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 51.110652663165794,
    },
    "grok-3-reason": {
        "id": "grok-3-reason",
        "name": "grok-3-reason",
        "model": "grok-3-reason",
        "provider": "xAI",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 62.81976744186046,
    },
    "o3-mini-2025-01-31": {
        "id": "o3-mini-2025-01-31",
        "name": "o3-mini-2025-01-31",
        "model": "o3-mini-2025-01-31",
        "provider": "Unknown",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 125.31410256410257,
    },
    "qwen3-235b-a22b": {
        "id": "qwen3-235b-a22b",
        "name": "qwen3-235b-a22b",
        "model": "qwen3-235b-a22b",
        "provider": "Alibaba",
        "maxLength": 0,
        "tokenLimit": 0,
        "context": 0,
        "success_rate": 100,
        "tps": 25.846153846153847,
    },
}

class Liaobots(AsyncGeneratorProvider, ProviderModelMixin):
    url = "https://liaobots.work"
    working = True
    supports_message_history = True
    supports_system_message = True
    
    default_model = "grok-3"
    models = list(models.keys())
    model_aliases = {
        # Anthropic
        "claude-3.5-sonnet": "claude-3-5-sonnet-20241022",
        "claude-3.5-sonnet": "claude-3-5-sonnet-20241022-t",
        "claude-3.7-sonnet": "claude-3-7-sonnet-20250219",
        "claude-3.7-sonnet": "claude-3-7-sonnet-20250219-t",
        
        # DeepSeek
        #"deepseek-v3": "deepseek-v3",
        
        # Google
        "gemini-1.0-pro": "gemini-1.0-pro-latest-123",
        "gemini-2.0-flash": "gemini-2.0-flash-exp",
        "gemini-2.0-flash-thinking": "gemini-2.0-flash-thinking-exp",
        "gemini-2.5-flash": "gemini-2.5-flash-preview-04-17",
        "gemini-2.5-pro": "gemini-2.5-pro-official",
        "gemini-2.5-pro": "gemini-2.5-pro-preview-03-25",
        "gemini-2.5-pro": "gemini-2.5-pro-preview-05-06",
        
        # OpenAI
        "gpt-4-turbo": "gpt-4-turbo-2024-04-09",
        "gpt-4.1-mini": "gpt-4.1-mini-2025-04-14",
        "gpt-4": "gpt-4o-2024-11-20",
        "gpt-4o": "gpt-4o-2024-11-20",
        "gpt-4o-mini": "gpt-4o-mini-2024-07-18",
        
        # xAI
        "grok-3-reason": "grok-3-reason",
        "o3-mini": "o3-mini-2025-01-31",
        "qwen-3-235b": "qwen3-235b-a22b",
    }
    
    _auth_code = None
    _cookie_jar = None

    @classmethod
    def is_supported(cls, model: str) -> bool:
        """
        Check if the given model is supported.
        """
        return model in models or model in cls.model_aliases

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        connector: BaseConnector = None,
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        
        headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://liaobots.work",
            "priority": "u=1, i",
            "referer": "https://liaobots.work/en",
            "sec-ch-ua": "\"Chromium\";v=\"135\", \"Not-A.Brand\";v=\"8\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Linux\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36"
        }
        
        async with ClientSession(
            headers=headers,
            cookie_jar=cls._cookie_jar,
            connector=get_connector(connector, proxy, True)
        ) as session:
            # First, get a valid auth code
            await cls.get_auth_code(session)
            
            # Create conversation ID
            conversation_id = str(uuid.uuid4())
            
            # Prepare request data
            data = {
                "conversationId": conversation_id,
                "models": [{
                    "modelId": model,
                    "provider": models[model]["provider"]
                }],
                "search": "false",
                "messages": messages,
                "key": "",
                "prompt": kwargs.get("system_message", "你是 {{model}}，一个由 {{provider}} 训练的大型语言模型，请仔细遵循用户的指示。")
            }
            
            # Try to make the chat request
            try:
                # Make the chat request with the current auth code
                async with session.post(
                    f"{cls.url}/api/chat",
                    json=data,
                    headers={"x-auth-code": cls._auth_code},
                    ssl=False
                ) as response:
                    # Check if we got a streaming response
                    content_type = response.headers.get("Content-Type", "")
                    if "text/event-stream" in content_type:
                        async for line in response.content:
                            if line.startswith(b"data: "):
                                try:
                                    response_data = json.loads(line[6:])
                                    
                                    # Check for error response
                                    if response_data.get("error") is True:
                                        # Raise RateLimitError for payment required or other errors
                                        if "402" in str(response_data.get("res_status", "")):
                                            raise RateLimitError("This model requires payment or credits")
                                        else:
                                            error_msg = response_data.get('message', 'Unknown error')
                                            raise RateLimitError(f"Error: {error_msg}")
                                    
                                    # Process normal response
                                    if response_data.get("role") == "assistant" and "content" in response_data:
                                        content = response_data.get("content")
                                        yield content
                                except json.JSONDecodeError:
                                    continue
                    else:
                        # Not a streaming response, might be an error or HTML
                        response_text = await response.text()
                        
                        # If we got HTML, we need to bypass CAPTCHA
                        if response_text.startswith("<!DOCTYPE html>"):
                            await cls.bypass_captcha(session)
                            
                            # Get a fresh auth code
                            await cls.get_auth_code(session)
                            
                            # Try the request again
                            async with session.post(
                                f"{cls.url}/api/chat",
                                json=data,
                                headers={"x-auth-code": cls._auth_code},
                                ssl=False
                            ) as response2:
                                # Check if we got a streaming response
                                content_type = response2.headers.get("Content-Type", "")
                                if "text/event-stream" in content_type:
                                    async for line in response2.content:
                                        if line.startswith(b"data: "):
                                            try:
                                                response_data = json.loads(line[6:])
                                                
                                                # Check for error response
                                                if response_data.get("error") is True:
                                                    # Raise RateLimitError for payment required or other errors
                                                    if "402" in str(response_data.get("res_status", "")):
                                                        raise RateLimitError("This model requires payment or credits")
                                                    else:
                                                        error_msg = response_data.get('message', 'Unknown error')
                                                        raise RateLimitError(f"Error: {error_msg}")
                                                
                                                # Process normal response
                                                if response_data.get("role") == "assistant" and "content" in response_data:
                                                    content = response_data.get("content")
                                                    yield content
                                            except json.JSONDecodeError:
                                                continue
                                else:
                                    raise RateLimitError("Failed to get streaming response")
                        else:
                            raise RateLimitError("Failed to connect to the service")
            except Exception as e:
                # If it's already a RateLimitError, re-raise it
                if isinstance(e, RateLimitError):
                    raise
                # Otherwise, wrap it in a RateLimitError
                raise RateLimitError(f"Error processing request: {str(e)}")

    @classmethod
    async def bypass_captcha(cls, session: ClientSession) -> None:
        """
        Bypass the CAPTCHA verification by directly making the recaptcha API request.
        """
        try:
            # First, try the direct recaptcha API request
            async with session.post(
                f"{cls.url}/recaptcha/api/login",
                json={"token": "abcdefghijklmnopqrst"},
                ssl=False
            ) as response:
                if response.status == 200:
                    try:
                        response_text = await response.text()
                        
                        # Try to parse as JSON
                        try:
                            response_data = json.loads(response_text)
                            
                            # Check if we got a successful response
                            if response_data.get("code") == 200:
                                cls._cookie_jar = session.cookie_jar
                        except json.JSONDecodeError:
                            pass
                    except Exception:
                        pass
        except Exception:
            pass

    @classmethod
    async def get_auth_code(cls, session: ClientSession) -> None:
        """
        Get a valid auth code by sending a request with an empty authcode.
        """
        try:
            # Send request with empty authcode to get a new one
            auth_request_data = {
                "authcode": "",
                "recommendUrl": "https://liaobots.work/zh"
            }
            
            async with session.post(
                f"{cls.url}/api/user",
                json=auth_request_data,
                ssl=False
            ) as response:
                if response.status == 200:
                    response_text = await response.text()
                    
                    try:
                        response_data = json.loads(response_text)
                        
                        if "authCode" in response_data:
                            cls._auth_code = response_data["authCode"]
                            cls._cookie_jar = session.cookie_jar
                            return
                    except json.JSONDecodeError:
                        # If we got HTML, it might be the CAPTCHA page
                        if response_text.startswith("<!DOCTYPE html>"):
                            await cls.bypass_captcha(session)
                            
                            # Try again after bypassing CAPTCHA
                            async with session.post(
                                f"{cls.url}/api/user",
                                json=auth_request_data,
                                ssl=False
                            ) as response2:
                                if response2.status == 200:
                                    response_text2 = await response2.text()
                                    
                                    try:
                                        response_data2 = json.loads(response_text2)
                                        
                                        if "authCode" in response_data2:
                                            cls._auth_code = response_data2["authCode"]
                                            cls._cookie_jar = session.cookie_jar
                                            return
                                    except json.JSONDecodeError:
                                        pass
        except Exception:
            pass
            
        # If we're here, we couldn't get a valid auth code
        # Set a default one as a fallback
        cls._auth_code = "DvS3A5GTE9f0D"  # Fallback to one of the provided auth codes



from __future__ import annotations

from aiohttp import ClientSession
import asyncio
import random
from typing import Optional
from ...typing import AsyncResult, Messages
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ...providers.response import ImageResponse

class Prodia(AsyncGeneratorProvider, ProviderModelMixin):
    url = "https://app.prodia.com"
    api_endpoint = "https://api.prodia.com/generate"
    
    working = False
    
    default_model = 'absolutereality_v181.safetensors [3d9d4d2b]'
    default_image_model = default_model
    image_models = [
        '3Guofeng3_v34.safetensors [50f420de]',
        'absolutereality_V16.safetensors [37db0fc3]',
        default_image_model,
        'amIReal_V41.safetensors [0a8a2e61]',
        'analog-diffusion-1.0.ckpt [9ca13f02]',
        'aniverse_v30.safetensors [579e6f85]',
        'anythingv3_0-pruned.ckpt [2700c435]',
        'anything-v4.5-pruned.ckpt [65745d25]',
        'anythingV5_PrtRE.safetensors [893e49b9]',
        'AOM3A3_orangemixs.safetensors [9600da17]',
        'blazing_drive_v10g.safetensors [ca1c1eab]',
        'breakdomain_I2428.safetensors [43cc7d2f]',
        'breakdomain_M2150.safetensors [15f7afca]',
        'cetusMix_Version35.safetensors [de2f2560]',
        'childrensStories_v13D.safetensors [9dfaabcb]',
        'childrensStories_v1SemiReal.safetensors [a1c56dbb]',
        'childrensStories_v1ToonAnime.safetensors [2ec7b88b]',
        'Counterfeit_v30.safetensors [9e2a8f19]',
        'cuteyukimixAdorable_midchapter3.safetensors [04bdffe6]',
        'cyberrealistic_v33.safetensors [82b0d085]',
        'dalcefo_v4.safetensors [425952fe]',
        'deliberate_v2.safetensors [10ec4b29]',
        'deliberate_v3.safetensors [afd9d2d4]',
        'dreamlike-anime-1.0.safetensors [4520e090]',
        'dreamlike-diffusion-1.0.safetensors [5c9fd6e0]',
        'dreamlike-photoreal-2.0.safetensors [fdcf65e7]',
        'dreamshaper_6BakedVae.safetensors [114c8abb]',
        'dreamshaper_7.safetensors [5cf5ae06]',
        'dreamshaper_8.safetensors [9d40847d]',
        'edgeOfRealism_eorV20.safetensors [3ed5de15]',
        'EimisAnimeDiffusion_V1.ckpt [4f828a15]',
        'elldreths-vivid-mix.safetensors [342d9d26]',
        'epicphotogasm_xPlusPlus.safetensors [1a8f6d35]',
        'epicrealism_naturalSinRC1VAE.safetensors [90a4c676]',
        'epicrealism_pureEvolutionV3.safetensors [42c8440c]',
        'ICantBelieveItsNotPhotography_seco.safetensors [4e7a3dfd]',
        'indigoFurryMix_v75Hybrid.safetensors [91208cbb]',
        'juggernaut_aftermath.safetensors [5e20c455]',
        'lofi_v4.safetensors [ccc204d6]',
        'lyriel_v16.safetensors [68fceea2]',
        'majicmixRealistic_v4.safetensors [29d0de58]',
        'mechamix_v10.safetensors [ee685731]',
        'meinamix_meinaV9.safetensors [2ec66ab0]',
        'meinamix_meinaV11.safetensors [b56ce717]',
        'neverendingDream_v122.safetensors [f964ceeb]',
        'openjourney_V4.ckpt [ca2f377f]',
        'pastelMixStylizedAnime_pruned_fp16.safetensors [793a26e8]',
        'portraitplus_V1.0.safetensors [1400e684]',
        'protogenx34.safetensors [5896f8d5]',
        'Realistic_Vision_V1.4-pruned-fp16.safetensors [8d21810b]',
        'Realistic_Vision_V2.0.safetensors [79587710]',
        'Realistic_Vision_V4.0.safetensors [29a7afaa]',
        'Realistic_Vision_V5.0.safetensors [614d1063]',
        'Realistic_Vision_V5.1.safetensors [a0f13c83]',
        'redshift_diffusion-V10.safetensors [1400e684]',
        'revAnimated_v122.safetensors [3f4fefd9]',
        'rundiffusionFX25D_v10.safetensors [cd12b0ee]',
        'rundiffusionFX_v10.safetensors [cd4e694d]',
        'sdv1_4.ckpt [7460a6fa]',
        'v1-5-pruned-emaonly.safetensors [d7049739]',
        'v1-5-inpainting.safetensors [21c7ab71]',
        'shoninsBeautiful_v10.safetensors [25d8c546]',
        'theallys-mix-ii-churned.safetensors [5d9225a4]',
        'timeless-1.0.ckpt [7c4971d4]',
        'toonyou_beta6.safetensors [980f6b15]'
    ]
    models = [*image_models]

    @classmethod
    def get_model(cls, model: str) -> str:
        if model in cls.models:
            return model
        elif model in cls.model_aliases:
            return cls.model_aliases[model]
        else:
            return cls.default_model

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        negative_prompt: str = "",
        steps: str = 20, # 1-25
        cfg: str = 7, # 0-20
        seed: Optional[int] = None,
        sampler: str = "DPM++ 2M Karras", # "Euler", "Euler a", "Heun", "DPM++ 2M Karras", "DPM++ SDE Karras", "DDIM"
        aspect_ratio: str = "square", # "square", "portrait", "landscape"
        **kwargs
    ) -> AsyncResult:
        model = cls.get_model(model)
        
        if seed is None:
            seed = random.randint(0, 10000)
        
        headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "origin": cls.url,
            "referer": f"{cls.url}/",
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36"
        }
        
        async with ClientSession(headers=headers) as session:
            prompt = messages[-1]['content'] if messages else ""
            
            params = {
                "new": "true",
                "prompt": prompt,
                "model": model,
                "negative_prompt": negative_prompt,
                "steps": steps,
                "cfg": cfg,
                "seed": seed,
                "sampler": sampler,
                "aspect_ratio": aspect_ratio
            }
            
            async with session.get(cls.api_endpoint, params=params, proxy=proxy) as response:
                response.raise_for_status()
                job_data = await response.json()
                job_id = job_data["job"]
                
                image_url = await cls._poll_job(session, job_id, proxy)
                yield ImageResponse(image_url, alt=prompt)

    @classmethod
    async def _poll_job(cls, session: ClientSession, job_id: str, proxy: str, max_attempts: int = 30, delay: int = 2) -> str:
        for _ in range(max_attempts):
            async with session.get(f"https://api.prodia.com/job/{job_id}", proxy=proxy) as response:
                response.raise_for_status()
                job_status = await response.json()

                if job_status["status"] == "succeeded":
                    return f"https://images.prodia.xyz/{job_id}.png"
                elif job_status["status"] == "failed":
                    raise Exception("Image generation failed")

            await asyncio.sleep(delay)

        raise Exception("Timeout waiting for image generation")



from __future__ import annotations

import argparse
from argparse import ArgumentParser

from .client import get_parser, run_client_args
from ..requests import BrowserConfig
from ..gui.run import gui_parser, run_gui_args
from ..config import DEFAULT_PORT, DEFAULT_TIMEOUT, DEFAULT_STREAM_TIMEOUT
from .. import Provider
from .. import cookies

def get_api_parser():
    api_parser = ArgumentParser(description="Run the API and GUI")
    api_parser.add_argument("--bind", default=None, help=f"The bind string. (Default: 0.0.0.0:{DEFAULT_PORT})")
    api_parser.add_argument("--port", "-p", default=None, help=f"Change the port of the server. (Default: {DEFAULT_PORT})")
    api_parser.add_argument("--debug", "-d", action="store_true", help="Enable verbose logging.")
    api_parser.add_argument("--gui", "-g", default=None, action="store_true", help="(deprecated)")
    api_parser.add_argument("--no-gui", "-ng", default=False, action="store_true", help="Start without the gui.")
    api_parser.add_argument("--model", default=None, help="Default model for chat completion. (incompatible with --reload and --workers)")
    api_parser.add_argument("--provider", choices=[provider.__name__ for provider in Provider.__providers__ if provider.working],
                            default=None, help="Default provider for chat completion. (incompatible with --reload and --workers)")
    api_parser.add_argument("--media-provider", choices=[provider.__name__ for provider in Provider.__providers__ if provider.working and bool(getattr(provider, "image_models", False))],
                            default=None, help="Default provider for image generation. (incompatible with --reload and --workers)"),
    api_parser.add_argument("--proxy", default=None, help="Default used proxy. (incompatible with --reload and --workers)")
    api_parser.add_argument("--workers", type=int, default=None, help="Number of workers.")
    api_parser.add_argument("--disable-colors", action="store_true", help="Don't use colors.")
    api_parser.add_argument("--ignore-cookie-files", action="store_true", help="Don't read .har and cookie files. (incompatible with --reload and --workers)")
    api_parser.add_argument("--g4f-api-key", type=str, default=None, help="Sets an authentication key for your API. (incompatible with --reload and --workers)")
    api_parser.add_argument("--ignored-providers", nargs="+", choices=[provider.__name__ for provider in Provider.__providers__ if provider.working],
                            default=[], help="List of providers to ignore when processing request. (incompatible with --reload and --workers)")
    api_parser.add_argument("--cookie-browsers", nargs="+", choices=[browser.__name__ for browser in cookies.BROWSERS],
                            default=[], help="List of browsers to access or retrieve cookies from. (incompatible with --reload and --workers)")
    api_parser.add_argument("--reload", action="store_true", help="Enable reloading.")
    api_parser.add_argument("--demo", action="store_true", help="Enable demo mode.")
    api_parser.add_argument("--timeout", type=int, default=DEFAULT_TIMEOUT, help="Default timeout for requests in seconds. (incompatible with --reload and --workers)")
    api_parser.add_argument("--stream-timeout", type=int, default=DEFAULT_STREAM_TIMEOUT, help="Default timeout for streaming requests in seconds. (incompatible with --reload and --workers)")
    api_parser.add_argument("--ssl-keyfile", type=str, default=None, help="Path to SSL key file for HTTPS.")
    api_parser.add_argument("--ssl-certfile", type=str, default=None, help="Path to SSL certificate file for HTTPS.")
    api_parser.add_argument("--log-config", type=str, default=None, help="Custom log config.")
    api_parser.add_argument("--browser-port", type=int, help="Port for the browser automation tool.")
    api_parser.add_argument("--browser-host", type=str, default="127.0.0.1", help="Host for the browser automation tool.")

    return api_parser

def run_api_args(args):
    from g4f11.api import AppConfig, run_api

    AppConfig.set_config(
        ignore_cookie_files=args.ignore_cookie_files,
        ignored_providers=args.ignored_providers,
        g4f_api_key=args.g4f_api_key,
        provider=args.provider,
        media_provider=args.media_provider,
        proxy=args.proxy,
        model=args.model,
        gui=not args.no_gui,
        demo=args.demo,
        timeout=args.timeout,
        stream_timeout=args.stream_timeout
    )

    if args.browser_port:
        BrowserConfig.port = args.browser_port
        BrowserConfig.host = args.browser_host
    if args.cookie_browsers:
        cookies.BROWSERS = [cookies[browser] for browser in args.cookie_browsers]

    run_api(
        bind=args.bind,
        port=args.port,
        debug=args.debug,
        workers=args.workers,
        use_colors=not args.disable_colors,
        reload=args.reload,
        ssl_keyfile=args.ssl_keyfile,
        ssl_certfile=args.ssl_certfile,
        log_config=args.log_config,
    )

def main():
    parser = argparse.ArgumentParser(description="Run gpt4free", exit_on_error=False)
    subparsers = parser.add_subparsers(dest="mode", help="Mode to run the g4f in.")
    subparsers.add_parser("api", parents=[get_api_parser()], add_help=False)
    subparsers.add_parser("gui", parents=[gui_parser()], add_help=False)
    subparsers.add_parser("client", parents=[get_parser()], add_help=False)

    try:
        args = parser.parse_args()
        if args.mode == "api":
            run_api_args(args)
        elif args.mode == "gui":
            run_gui_args(args)
        elif args.mode == "client":
            run_client_args(args)
        else:
            raise argparse.ArgumentError(None, "No valid mode specified. Use 'api', 'gui', or 'client'.")
    except argparse.ArgumentError:
        try:
            run_client_args(get_parser(exit_on_error=False).parse_args(), exit_on_error=False)
        except argparse.ArgumentError:
            run_api_args(get_api_parser().parse_args())


from __future__ import annotations

import os
import time
import random
import string
import asyncio
import aiohttp
import base64
from typing import Union, AsyncIterator, Iterator, Awaitable, Optional

from ..image.copy_images import copy_media, get_media_dir
from ..typing import Messages, ImageType
from ..providers.types import ProviderType, BaseRetryProvider
from ..providers.response import *
from ..errors import NoMediaResponseError
from ..providers.retry_provider import IterListProvider
from ..providers.asyncio import to_sync_generator
from ..providers.any_provider import AnyProvider
from ..Provider import OpenaiAccount, PollinationsImage
from ..tools.run_tools import async_iter_run_tools, iter_run_tools
from .stubs import ChatCompletion, ChatCompletionChunk, Image, ImagesResponse, UsageModel, ToolCallModel, ClientResponse
from .models import ClientModels
from .types import IterResponse, Client as BaseClient
from .service import convert_to_provider
from .helper import find_stop, filter_json, filter_none, safe_aclose
from .. import debug

ChatCompletionResponseType = Iterator[Union[ChatCompletion, ChatCompletionChunk, BaseConversation]]
AsyncChatCompletionResponseType = AsyncIterator[Union[ChatCompletion, ChatCompletionChunk, BaseConversation]]

try:
    anext # Python 3.8+
except NameError:
    async def anext(aiter):
        try:
            return await aiter.__anext__()
        except StopAsyncIteration:
            raise StopIteration

def add_chunk(content, chunk):
    if content == "" and isinstance(chunk, (MediaResponse, AudioResponse)):
        content = chunk
    elif not isinstance(chunk, (Reasoning, ToolCalls)):
        content = str(content) + str(chunk)
    return content

def resolve_media(kwargs: dict, image = None, image_name: str = None) -> None:
    if image is not None:
        kwargs["media"] = [(image, getattr(image, "name", image_name))]
    elif "images" in kwargs:
        kwargs["media"] = kwargs.pop("images")
    if kwargs.get("media") is None:
        kwargs.pop("media", None)
    elif not isinstance(kwargs["media"], list):
        kwargs["media"] = [kwargs["media"]]
    for idx, media in enumerate(kwargs.get("media", [])):
        if not isinstance(media, (list, tuple)):
            kwargs["media"][idx] = (media, getattr(media, "name", None))

# Synchronous iter_response function
def iter_response(
    response: Union[Iterator[Union[str, ResponseType]]],
    stream: bool,
    response_format: Optional[dict] = None,
    max_tokens: Optional[int] = None,
    stop: Optional[list[str]] = None
) -> ChatCompletionResponseType:
    content = ""
    reasoning = []
    finish_reason = None
    tool_calls = None
    usage = None
    provider: ProviderInfo = None
    conversation: JsonConversation = None
    completion_id = ''.join(random.choices(string.ascii_letters + string.digits, k=28))
    idx = 0

    if hasattr(response, '__aiter__'):
        response = to_sync_generator(response)

    for chunk in response:
        if isinstance(chunk, FinishReason):
            finish_reason = chunk.reason
            continue
        elif isinstance(chunk, JsonConversation):
            conversation = chunk
            continue
        elif isinstance(chunk, ToolCalls):
            if not stream:
                tool_calls = chunk.get_list()
                continue
        elif isinstance(chunk, Usage):
            usage = chunk
            continue
        elif isinstance(chunk, ProviderInfo):
            provider = chunk
            continue
        elif isinstance(chunk, Reasoning):
            reasoning.append(chunk)
        elif isinstance(chunk, HiddenResponse):
            continue
        elif isinstance(chunk, Exception):
            continue
        elif not chunk:
            continue

        content = add_chunk(content, chunk)
        idx += 1

        if max_tokens is not None and idx >= max_tokens:
            finish_reason = "length"

        first, content, chunk = find_stop(stop, content, chunk if stream else None)

        if first != -1:
            finish_reason = "stop"

        if stream:
            chunk = ChatCompletionChunk.model_construct(chunk, None, completion_id, int(time.time()))
            if provider is not None:
                chunk.provider = provider.name
                chunk.model = provider.model
            yield chunk

        if finish_reason is not None:
            break

    if usage is None:
        usage = UsageModel.model_construct(completion_tokens=idx, total_tokens=idx)
    else:
        usage = UsageModel.model_construct(**usage.get_dict())

    finish_reason = "stop" if finish_reason is None else finish_reason

    if stream:
        chat_completion = ChatCompletionChunk.model_construct(
            None, finish_reason, completion_id, int(time.time()), usage=usage
        )
    else:
        if response_format is not None and "type" in response_format:
            if response_format["type"] == "json_object":
                content = filter_json(content)
        chat_completion = ChatCompletion.model_construct(
            content, finish_reason, completion_id, int(time.time()), usage=usage,
            **(filter_none(
                tool_calls=[ToolCallModel.model_construct(**tool_call) for tool_call in tool_calls]
            ) if tool_calls is not None else {}),
            conversation=None if conversation is None else conversation.get_dict(),
            reasoning=reasoning if reasoning else None
        )
    if provider is not None:
        chat_completion.provider = provider.name
        chat_completion.model = provider.model
    yield chat_completion

# Synchronous iter_append_model_and_provider function
def iter_append_model_and_provider(response: ChatCompletionResponseType, last_model: str, last_provider: ProviderType) -> ChatCompletionResponseType:
    if isinstance(last_provider, BaseRetryProvider):
        yield from response
        return
    for chunk in response:
        if isinstance(chunk, (ChatCompletion, ChatCompletionChunk)):
            if chunk.provider is None and last_provider is not None:
                chunk.model = getattr(last_provider, "last_model", last_model)
                chunk.provider = last_provider.__name__
        yield chunk

async def async_iter_response(
    response: AsyncIterator[Union[str, ResponseType]],
    stream: bool,
    response_format: Optional[dict] = None,
    max_tokens: Optional[int] = None,
    stop: Optional[list[str]] = None
) -> AsyncChatCompletionResponseType:
    content = ""
    reasoning = []
    finish_reason = None
    completion_id = ''.join(random.choices(string.ascii_letters + string.digits, k=28))
    idx = 0
    tool_calls = None
    usage = None
    provider: ProviderInfo = None
    conversation: JsonConversation = None

    try:
        async for chunk in response:
            if isinstance(chunk, FinishReason):
                finish_reason = chunk.reason
                continue
            elif isinstance(chunk, JsonConversation):
                conversation = chunk
                continue
            elif isinstance(chunk, ToolCalls):
                if not stream:
                    tool_calls = chunk.get_list()
                    continue
            elif isinstance(chunk, Usage):
                usage = chunk
                continue
            elif isinstance(chunk, ProviderInfo):
                provider = chunk
                continue
            elif isinstance(chunk, Reasoning) and not stream:
                reasoning.append(chunk)
            elif isinstance(chunk, HiddenResponse):
                continue
            elif isinstance(chunk, Exception):
                continue
            elif not chunk:
                continue

            content = add_chunk(content, chunk)
            idx += 1

            if max_tokens is not None and idx >= max_tokens:
                finish_reason = "length"

            first, content, chunk = find_stop(stop, content, chunk if stream else None)

            if first != -1:
                finish_reason = "stop"

            if stream:
                chunk = ChatCompletionChunk.model_construct(chunk, None, completion_id, int(time.time()))
                if provider is not None:
                    chunk.provider = provider.name
                    chunk.model = provider.model
                yield chunk

            if finish_reason is not None:
                break

        finish_reason = "stop" if finish_reason is None else finish_reason

        if usage is None:
            usage = UsageModel.model_construct(completion_tokens=idx, total_tokens=idx)
        else:
            usage = UsageModel.model_construct(**usage.get_dict())

        if stream:
            chat_completion = ChatCompletionChunk.model_construct(
                None, finish_reason, completion_id, int(time.time()), usage=usage, conversation=conversation
            )
        else:
            if response_format is not None and "type" in response_format:
                if response_format["type"] == "json_object":
                    content = filter_json(content)
            chat_completion = ChatCompletion.model_construct(
                content, finish_reason, completion_id, int(time.time()), usage=usage,
                **(filter_none(
                    tool_calls=[ToolCallModel.model_construct(**tool_call) for tool_call in tool_calls]
                ) if tool_calls else {}),
                conversation=conversation,
                reasoning=reasoning if reasoning else None
            )
        if provider is not None:
            chat_completion.provider = provider.name
            chat_completion.model = provider.model
        yield chat_completion
    finally:
        await safe_aclose(response)

async def async_response(
    response: AsyncIterator[Union[str, ResponseType]]
) -> ClientResponse:
    content = ""
    response_id = ''.join(random.choices(string.ascii_letters + string.digits, k=28))
    idx = 0
    usage = None
    provider: ProviderInfo = None
    conversation: JsonConversation = None

    async for chunk in response:
        if isinstance(chunk, FinishReason):
            continue
        elif isinstance(chunk, JsonConversation):
            conversation = chunk
            continue
        elif isinstance(chunk, ToolCalls):
            continue
        elif isinstance(chunk, Usage):
            usage = chunk
            continue
        elif isinstance(chunk, ProviderInfo):
            provider = chunk
            continue
        elif isinstance(chunk, HiddenResponse):
            continue
        elif isinstance(chunk, Exception):
            continue

        content = add_chunk(content, chunk)
        if not content:
            continue
        idx += 1

    if usage is None:
        usage = UsageModel.model_construct(completion_tokens=idx, total_tokens=idx)
    else:
        usage = UsageModel.model_construct(**usage.get_dict())

    response = ClientResponse.model_construct(
        content, response_id, int(time.time()), usage=usage, conversation=conversation
    )
    if provider is not None:
        response.provider = provider.name
        response.model = provider.model
    return response

async def async_iter_append_model_and_provider(
        response: AsyncChatCompletionResponseType,
        last_model: str,
        last_provider: ProviderType
    ) -> AsyncChatCompletionResponseType:
    try:
        if isinstance(last_provider, BaseRetryProvider):
            async for chunk in response:
                yield chunk
            return
        async for chunk in response:
            if isinstance(chunk, (ChatCompletion, ChatCompletionChunk)):
                if chunk.provider is None and last_provider is not None:
                    chunk.model = getattr(last_provider, "last_model", last_model)
                    chunk.provider = last_provider.__name__
            yield chunk
    finally:
        await safe_aclose(response)

class Client(BaseClient):
    def __init__(
        self,
        provider: Optional[ProviderType] = None,
        media_provider: Optional[ProviderType] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.chat: Chat = Chat(self, provider)
        if media_provider is None:
            media_provider = kwargs.get("image_provider", provider)
        self.models: ClientModels = ClientModels(self, provider, media_provider)
        self.images: Images = Images(self, media_provider)
        self.media: Images = self.images

class Completions:
    def __init__(self, client: Client, provider: Optional[ProviderType] = None):
        self.client: Client = client
        self.provider: ProviderType = provider

    def create(
        self,
        messages: Messages,
        model: str = "",
        provider: Optional[ProviderType] = None,
        stream: Optional[bool] = False,
        proxy: Optional[str] = None,
        image: Optional[ImageType] = None,
        image_name: Optional[str] = None,
        response_format: Optional[dict] = None,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[list[str], str]] = None,
        api_key: Optional[str] = None,
        ignore_stream: Optional[bool] = False,
        **kwargs
    ) -> ChatCompletion:
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        resolve_media(kwargs, image, image_name)
        if hasattr(model, "name"):
            model = model.get_long_name()
        if provider is None:
            provider = self.provider
            if provider is None:
                provider = AnyProvider
        if isinstance(provider, str):
            provider = convert_to_provider(provider)
        stop = [stop] if isinstance(stop, str) else stop
        if ignore_stream:
            kwargs["ignore_stream"] = True

        response = iter_run_tools(
            provider,
            model=model,
            messages=messages,
            stream=stream,
            **filter_none(
                proxy=self.client.proxy if proxy is None else proxy,
                max_tokens=max_tokens,
                stop=stop,
                api_key=self.client.api_key if api_key is None else api_key
            ),
            **kwargs
        )

        response = iter_response(response, stream, response_format, max_tokens, stop)
        response = iter_append_model_and_provider(response, model, provider)
        if stream:
            return response
        else:
            return next(response)

    def stream(
        self,
        messages: Messages,
        model: str = "",
        **kwargs
    ) -> IterResponse:
        return self.create(messages, model, stream=True, **kwargs)

class Chat:
    completions: Completions

    def __init__(self, client: Client, provider: Optional[ProviderType] = None):
        self.completions = Completions(client, provider)

class Images:
    def __init__(self, client: Client, provider: Optional[ProviderType] = None):
        self.client: Client = client
        self.provider: Optional[ProviderType] = provider

    def generate(
        self,
        prompt: str,
        model: str = None,
        provider: Optional[ProviderType] = None,
        response_format: Optional[str] = None,
        proxy: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        """
        Synchronous generate method that runs the async_generate method in an event loop.
        """
        return asyncio.run(self.async_generate(prompt, model, provider, response_format, proxy, **kwargs))

    async def get_provider_handler(self, model: Optional[str], provider: Optional[ProviderType], default: ProviderType) -> ProviderType:
        if provider is None:
            provider_handler = self.provider
            if provider_handler is None:
                provider_handler = self.client.models.get(model, default)
        else:
            provider_handler = provider
        if isinstance(provider_handler, str):
            provider_handler = convert_to_provider(provider_handler)
        return provider_handler

    async def async_generate(
        self,
        prompt: str,
        model: Optional[str] = None,
        provider: Optional[ProviderType] = None,
        response_format: Optional[str] = None,
        proxy: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        provider_handler = await self.get_provider_handler(model, provider, PollinationsImage)
        provider_name = provider_handler.__name__ if hasattr(provider_handler, "__name__") else type(provider_handler).__name__
        if proxy is None:
            proxy = self.client.proxy
        if api_key is None:
            api_key = self.client.api_key
        error = None
        response = None
        if isinstance(provider_handler, IterListProvider):
            for provider in provider_handler.providers:
                try:
                    response = await self._generate_image_response(provider, provider.__name__, model, prompt, proxy=proxy, api_key=api_key, **kwargs)
                    if response is not None:
                        provider_name = provider.__name__
                        break
                except Exception as e:
                    error = e
                    debug.error(f"{provider.__name__} {type(e).__name__}: {e}")
        else:
            response = await self._generate_image_response(provider_handler, provider_name, model, prompt, proxy=proxy, api_key=api_key, **kwargs)
        if response is None:
            if error is not None:
                raise error
            raise NoMediaResponseError(f"No media response from {provider_name}")
        return await self._process_image_response(
            response,
            model,
            provider_name,
            kwargs.get("download_media", True),
            response_format,
            proxy
        )

    async def _generate_image_response(
        self,
        provider_handler: ProviderType,
        provider_name: str,
        model: str,
        prompt: str,
        prompt_prefix: str = "Generate a image: ",
        api_key: str = None,
        **kwargs
    ) -> MediaResponse:
        messages = [{"role": "user", "content": f"{prompt_prefix}{prompt}"}]
        items: list[MediaResponse] = []
        if isinstance(api_key, dict):
            api_key = api_key.get(provider_handler.get_parent())
        if hasattr(provider_handler, "create_async_generator"):
            async for item in provider_handler.create_async_generator(
                model,
                messages,
                stream=True,
                prompt=prompt,
                api_key=api_key,
                **kwargs
            ):
                if isinstance(item, (MediaResponse, AudioResponse)) and not isinstance(item, HiddenResponse):
                    items.append(item)
        elif hasattr(provider_handler, "create_completion"):
            for item in provider_handler.create_completion(
                model,
                messages,
                True,
                prompt=prompt,
                api_key=api_key,
                **kwargs
            ):
                if isinstance(item, (MediaResponse, AudioResponse)) and not isinstance(item, HiddenResponse):
                    items.append(item)
        else:
            raise ValueError(f"Provider {provider_name} does not support image generation")
        urls = []
        for item in items:
            if isinstance(item, AudioResponse):
                urls.append(item.to_uri())
            elif isinstance(item.urls, str):
                urls.append(item.urls)
            elif isinstance(item.urls, list):
                urls.extend(item.urls)
        if not urls:
            return None
        alt = getattr(items[0], "alt", "")
        return MediaResponse(urls, alt, items[0].options)

    def create_variation(
        self,
        image: ImageType,
        image_name: str = None,
        prompt: str = "Create a variation of this image",
        model: str = None,
        provider: Optional[ProviderType] = None,
        response_format: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        return asyncio.run(self.async_create_variation(
           image=image,
           image_name=image_name,
           prompt=prompt,
           model=model,
           provider=provider,
           response_format=response_format,
           **kwargs
        ))

    async def async_create_variation(
        self,
        *,
        image: ImageType,
        image_name: str = None,
        prompt: str = "Create a variation of this image",
        model: Optional[str] = None,
        provider: Optional[ProviderType] = None,
        response_format: Optional[str] = None,
        proxy: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        provider_handler = await self.get_provider_handler(model, provider, OpenaiAccount)
        provider_name = provider_handler.__name__ if hasattr(provider_handler, "__name__") else type(provider_handler).__name__
        if proxy is None:
            proxy = self.client.proxy
        resolve_media(kwargs, image, image_name)
        error = None
        response = None
        if isinstance(provider_handler, IterListProvider):
            for provider in provider_handler.providers:
                try:
                    response = await self._generate_image_response(provider, provider.__name__, model, prompt, **kwargs)
                    if response is not None:
                        provider_name = provider.__name__
                        break
                except Exception as e:
                    error = e
                    debug.error(f"{provider.__name__} {type(e).__name__}: {e}")
        else:
            response = await self._generate_image_response(provider_handler, provider_name, model, prompt, **kwargs)
        if response is None:
            if error is not None:
                raise error
            raise NoMediaResponseError(f"No media response from {provider_name}")
        return await self._process_image_response(
            response,
            model,
            provider_name,
            kwargs.get("download_media", True),
            response_format,
            proxy
        )

    async def _process_image_response(
        self,
        response: MediaResponse,
        model: str,
        provider: str,
        download_media: bool,
        response_format: Optional[str] = None,
        proxy: str = None
    ) -> ImagesResponse:
        if response_format == "url":
            # Return original URLs without saving locally
            images = [Image.model_construct(url=image, revised_prompt=response.alt) for image in response.get_list()]
        elif response_format == "b64_json":
            # Convert URLs directly to base64 without saving
            async def get_b64_from_url(url: str) -> Image:
                if url.startswith("/media/"):
                    with open(os.path.join(get_media_dir(), os.path.basename(url)), "rb") as f:
                        b64_data = base64.b64encode(f.read()).decode()
                        return Image.model_construct(b64_json=b64_data, revised_prompt=response.alt)
                async with aiohttp.ClientSession(cookies=response.get("cookies")) as session:
                    async with session.get(url, proxy=proxy) as resp:
                        if resp.status == 200:
                            b64_data = base64.b64encode(await resp.read()).decode()
                            return Image.model_construct(b64_json=b64_data, revised_prompt=response.alt)
                return Image.model_construct(url=url, revised_prompt=response.alt)
            images = await asyncio.gather(*[get_b64_from_url(image) for image in response.get_list()])
        else:
            # Save locally for None (default) case
            images = response.get_list()
            if download_media or response.get("cookies") or response.get("headers"):
                images = await copy_media(response.get_list(), response.get("cookies"), response.get("headers"), proxy, response.alt)
            images = [Image.model_construct(url=image, revised_prompt=response.alt) for image in images]

        return ImagesResponse.model_construct(
            created=int(time.time()),
            data=images,
            model=model,
            provider=provider
        )

class AsyncClient(BaseClient):
    def __init__(
        self,
        provider: Optional[ProviderType] = None,
        media_provider: Optional[ProviderType] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.chat: AsyncChat = AsyncChat(self, provider)
        if media_provider is None:
            media_provider = kwargs.get("image_provider", provider)
        self.models: ClientModels = ClientModels(self, provider, media_provider)
        self.images: AsyncImages = AsyncImages(self, media_provider)
        self.media: AsyncImages = self.images
        self.responses: AsyncResponses = AsyncResponses(self, provider)

class AsyncChat:
    completions: AsyncCompletions

    def __init__(self, client: AsyncClient, provider: Optional[ProviderType] = None):
        self.completions = AsyncCompletions(client, provider)

class AsyncCompletions:
    def __init__(self, client: AsyncClient, provider: Optional[ProviderType] = None):
        self.client: AsyncClient = client
        self.provider: ProviderType = provider

    def create(
        self,
        messages: Messages,
        model: str = "",
        provider: Optional[ProviderType] = None,
        stream: Optional[bool] = False,
        proxy: Optional[str] = None,
        image: Optional[ImageType] = None,
        image_name: Optional[str] = None,
        response_format: Optional[dict] = None,
        max_tokens: Optional[int] = None,
        stop: Optional[Union[list[str], str]] = None,
        api_key: Optional[str] = None,
        ignore_stream: Optional[bool] = False,
        **kwargs
    ) -> Awaitable[ChatCompletion]:
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        resolve_media(kwargs, image, image_name)
        if hasattr(model, "name"):
            model = model.get_long_name()
        if provider is None:
            provider = self.provider
            if provider is None:
                provider = AnyProvider
        if isinstance(provider, str):
            provider = convert_to_provider(provider)
        stop = [stop] if isinstance(stop, str) else stop
        if ignore_stream:
            kwargs["ignore_stream"] = True
            
        response = async_iter_run_tools(
            provider,
            model=model,
            messages=messages,
            stream=stream,
            **filter_none(
                proxy=self.client.proxy if proxy is None else proxy,
                max_tokens=max_tokens,
                stop=stop,
                api_key=self.client.api_key if api_key is None else api_key
            ),
            **kwargs
        )

        response = async_iter_response(response, stream, response_format, max_tokens, stop)
        response = async_iter_append_model_and_provider(response, model, provider)

        if stream:
            return response
        else:
            return anext(response)

    def stream(
        self,
        messages: Messages,
        model: str = "",
        **kwargs
    ) -> AsyncIterator[ChatCompletionChunk]:
        return self.create(messages, model, stream=True, **kwargs)

class AsyncImages(Images):
    def __init__(self, client: AsyncClient, provider: Optional[ProviderType] = None):
        self.client: AsyncClient = client
        self.provider: Optional[ProviderType] = provider

    async def generate(
        self,
        prompt: str,
        model: Optional[str] = None,
        provider: Optional[ProviderType] = None,
        response_format: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        return await self.async_generate(prompt, model, provider, response_format, **kwargs)

    async def create_variation(
        self,
        image: ImageType,
        model: str = None,
        provider: ProviderType = None,
        response_format: Optional[str] = None,
        **kwargs
    ) -> ImagesResponse:
        return await self.async_create_variation(
           image=image, model=model, provider=provider, response_format=response_format, **kwargs
        )

class AsyncResponses():
    def __init__(self, client: AsyncClient, provider: Optional[ProviderType] = None):
        self.client: AsyncClient = client
        self.provider: ProviderType = provider

    async def create(
        self,
        input: str,
        model: str = "",
        provider: Optional[ProviderType] = None,
        instructions: Optional[str] = None,
        proxy: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs
    ) -> ClientResponse:
        if isinstance(input, str):
            input = [{"role": "user", "content": input}]
        if instructions is not None:
            input = [{"role": "developer", "content": instructions}] + input
        for idx, message in enumerate(input):
            if isinstance(message["content"], list):
                for key, value in enumerate(message["content"]):
                    if isinstance(value, dict) and value.get("type") == "input_text":
                        message["content"][key] = {"type": "text", "text": value.get("text")}
                    input[idx] = {"role": message["role"], "content": message["content"]}
        resolve_media(kwargs)
        if hasattr(model, "name"):
            model = model.get_long_name()
        if provider is None:
            provider = self.provider
            if provider is None:
                provider = AnyProvider
        if isinstance(provider, str):
            provider = convert_to_provider(provider)

        response = async_iter_run_tools(
            provider,
            model=model,
            messages=input,
            **filter_none(
                proxy=self.client.proxy if proxy is None else proxy,
                api_key=self.client.api_key if api_key is None else api_key
            ),
            **kwargs
        )

        return await async_response(response)



from __future__ import annotations

from .template import OpenaiTemplate
from ..config import DEFAULT_MODEL

class ApiAirforce(OpenaiTemplate):
    label = "Api.Airforce"
    url = "https://api.airforce"
    login_url = "https://panel.api.airforce/dashboard"
    api_base = "https://api.airforce/v1"
    working = True
    active_by_default = True
    use_image_size = True


from __future__ import annotations

import asyncio
import random
import json

from ..typing import AsyncResult, Messages
from ..requests import StreamSession, raise_for_status
from ..errors import ResponseError
from ..providers.response import FinishReason, Sources
from .base_provider import AsyncGeneratorProvider, ProviderModelMixin

API_URL = "https://www.perplexity.ai/socket.io/"
WS_URL = "wss://www.perplexity.ai/socket.io/"

class PerplexityLabs(AsyncGeneratorProvider, ProviderModelMixin):
    label = "Perplexity Labs"
    url = "https://labs.perplexity.ai"
    working = True
    active_by_default = True

    default_model = "r1-1776"
    models = [
        default_model,
        "sonar-pro",
        "sonar",
        "sonar-reasoning",
        "sonar-reasoning-pro",
    ]

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        **kwargs
    ) -> AsyncResult:
        # Enhanced headers to avoid Cloudflare detection
        user_agents = [
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        ]
        
        headers = {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
            "Cache-Control": "no-cache",
            "DNT": "1",
            "Origin": cls.url,
            "Pragma": "no-cache",
            "Referer": f"{cls.url}/",
            "Sec-Ch-Ua": '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Ch-Ua-Platform": '"Windows"',
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "none",
            "Sec-Fetch-User": "?1",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": random.choice(user_agents)
        }
        
        # Try different impersonation methods
        impersonate_options = ["chrome120", "chrome", "edge120", "safari"]
        
        for impersonate in impersonate_options:
            try:
                async with StreamSession(headers=headers, proxy=proxy, impersonate=impersonate) as session:
                    # Add a small delay to appear more human-like
                    await asyncio.sleep(random.uniform(0.5, 1.5))
                    
                    t = format(random.getrandbits(32), "08x")
                    async with session.get(
                        f"{API_URL}?EIO=4&transport=polling&t={t}"
                    ) as response:
                        await raise_for_status(response)
                        text = await response.text()
                    assert text.startswith("0")
                    sid = json.loads(text[1:])["sid"]
                    
                    # Add delay between requests
                    await asyncio.sleep(random.uniform(0.2, 0.8))
                    
                    post_data = '40{"jwt":"anonymous-ask-user"}'
                    async with session.post(
                        f"{API_URL}?EIO=4&transport=polling&t={t}&sid={sid}",
                        data=post_data
                    ) as response:
                        await raise_for_status(response)
                        assert await response.text() == "OK"
                    
                    # Add delay
                    await asyncio.sleep(random.uniform(0.2, 0.8))
                    
                    async with session.get(
                        f"{API_URL}?EIO=4&transport=polling&t={t}&sid={sid}",
                        data=post_data
                    ) as response:
                        await raise_for_status(response)
                        assert (await response.text()).startswith("40")
                    
                    # Add delay before WebSocket
                    await asyncio.sleep(random.uniform(0.5, 1.0))
                    
                    async with session.ws_connect(f"{WS_URL}?EIO=4&transport=websocket&sid={sid}", autoping=False) as ws:
                        await ws.send_str("2probe")
                        assert(await ws.receive_str() == "3probe")
                        await ws.send_str("5")
                        assert(await ws.receive_str() == "6")
                        
                        format_messages = []
                        last_is_assistant = False
                        for message in messages:
                            if message["role"] == "assistant":
                                if last_is_assistant:
                                    continue
                                last_is_assistant = True
                            else:
                                last_is_assistant = False
                            if isinstance(message["content"], str):
                                format_messages.append({
                                    "role": message["role"],
                                    "content": message["content"]
                                })
                        
                        message_data = {
                            "version": "2.18",
                            "source": "default",
                            "model": model,
                            "messages": format_messages
                        }
                        
                        await ws.send_str("42" + json.dumps(["perplexity_labs", message_data]))
                        last_message = 0
                        while True:
                            message = await ws.receive_str()
                            if message == "2":
                                if last_message == 0:
                                    raise RuntimeError("Unknown error")
                                await ws.send_str("3")
                                continue
                            try:
                                if not message.startswith("42"):
                                    continue
                                    
                                parsed_data = json.loads(message[2:])
                                message_type = parsed_data[0]
                                data = parsed_data[1]
                                
                                # Handle error responses
                                if message_type.endswith("_query_progress") and data.get("status") == "failed":
                                    error_message = data.get("text", "Unknown API error")
                                    raise ResponseError(f"API Error: {error_message}\n")
                                
                                # Handle normal responses
                                if "output" in data:
                                    if last_message == 0 and model == cls.default_model:
                                        yield "<think>"
                                    yield data["output"][last_message:]
                                    last_message = len(data["output"])
                                    if data["final"]:
                                        if data["citations"]:
                                            yield Sources(data["citations"])
                                        yield FinishReason("stop")
                                        break
                            except ResponseError as e:
                                # Re-raise ResponseError directly
                                raise e
                            except Exception as e:
                                raise ResponseError(f"Error processing message: {message}") from e
                # If we get here, the impersonation worked
                break
            except Exception as e:
                if impersonate == impersonate_options[-1]:
                    # Last attempt failed, re-raise the error
                    raise e
                # Try next impersonation method
                continue
            t = format(random.getrandbits(32), "08x")
            async with session.get(
                f"{API_URL}?EIO=4&transport=polling&t={t}"
            ) as response:
                await raise_for_status(response)
                text = await response.text()
            assert text.startswith("0")
            sid = json.loads(text[1:])["sid"]
            post_data = '40{"jwt":"anonymous-ask-user"}'
            async with session.post(
                f"{API_URL}?EIO=4&transport=polling&t={t}&sid={sid}",
                data=post_data
            ) as response:
                await raise_for_status(response)
                assert await response.text() == "OK"
            async with session.get(
                f"{API_URL}?EIO=4&transport=polling&t={t}&sid={sid}",
                data=post_data
            ) as response:
                await raise_for_status(response)
                assert (await response.text()).startswith("40")
            async with session.ws_connect(f"{WS_URL}?EIO=4&transport=websocket&sid={sid}", autoping=False) as ws:
                await ws.send_str("2probe")
                assert(await ws.receive_str() == "3probe")
                await ws.send_str("5")
                assert(await ws.receive_str() == "6")
                format_messages = []
                last_is_assistant = False
                for message in messages:
                    if message["role"] == "assistant":
                        if last_is_assistant:
                            continue
                        last_is_assistant = True
                    else:
                        last_is_assistant = False
                    if isinstance(message["content"], str):
                        format_messages.append({
                            "role": message["role"],
                            "content": message["content"]
                        })
                message_data = {
                    "version": "2.18",
                    "source": "default",
                    "model": model,
                    "messages": format_messages
                }
                await ws.send_str("42" + json.dumps(["perplexity_labs", message_data]))
                last_message = 0
                while True:
                    message = await ws.receive_str()
                    if message == "2":
                        if last_message == 0:
                            raise RuntimeError("Unknown error")
                        await ws.send_str("3")
                        continue
                    try:
                        if not message.startswith("42"):
                            continue
                            
                        parsed_data = json.loads(message[2:])
                        message_type = parsed_data[0]
                        data = parsed_data[1]
                        
                        # Handle error responses
                        if message_type.endswith("_query_progress") and data.get("status") == "failed":
                            error_message = data.get("text", "Unknown API error")
                            raise ResponseError(f"API Error: {error_message}\n")
                        
                        # Handle normal responses
                        if "output" in data:
                            if last_message == 0 and model == cls.default_model:
                                yield "<think>"
                            yield data["output"][last_message:]
                            last_message = len(data["output"])
                            if data["final"]:
                                if data["citations"]:
                                    yield Sources(data["citations"])
                                yield FinishReason("stop")
                                break
                    except ResponseError as e:
                        # Re-raise ResponseError directly
                        raise e
                    except Exception as e:
                        raise ResponseError(f"Error processing message: {message}") from e


if __name__ == "__main__":
    import asyncio
    
    async def test_perplexity_labs():
        # Test messages
        messages = [
            {"role": "user", "content": "What is the capital of France?"}
        ]
        
        print("Testing PerplexityLabs provider...")
        print("Sending message:", messages[0]["content"])
        print("Using model:", PerplexityLabs.default_model)
        
        try:
            async for response in PerplexityLabs.create_async_generator(
                model=PerplexityLabs.default_model,
                messages=messages
            ):
                if isinstance(response, str):
                    print(response, end="", flush=True)
                elif hasattr(response, 'citations'):  # Sources object
                    print(f"\nSources: {response.citations}")
                else:  # FinishReason
                    print(f"\nFinished with reason: {response}")
        except Exception as e:
            print(f"\nError: {str(e)}")
    
    # Run the test
    asyncio.run(test_perplexity_labs())



from __future__ import annotations

import os
import base64
from typing import Iterator, Union
from urllib.parse import urlparse
from pathlib import Path

from ..typing import Messages
from ..image import is_data_an_media, to_input_audio, is_valid_media, is_valid_audio, to_data_uri
from .files import get_bucket_dir, read_bucket

def render_media(bucket_id: str, name: str, url: str, as_path: bool = False, as_base64: bool = False, **kwargs) -> Union[str, Path]:
    if (as_base64 or as_path or url.startswith("/")):
        file = Path(get_bucket_dir(bucket_id, "thumbnail", name))
        if not file.exists():
            file = Path(get_bucket_dir(bucket_id, "media", name))
        if as_path:
            return file
        data = file.read_bytes()
        data_base64 = base64.b64encode(data).decode()
        if as_base64:
            return data_base64
        return f"data:{is_data_an_media(data, name)};base64,{data_base64}"
    return url

def render_part(part: dict) -> dict:
    if "type" in part:
        return part
    text = part.get("text")
    if text:
        return {
            "type": "text",
            "text": text
        }
    filename = part.get("name")
    if (filename is None):
        bucket_dir = Path(get_bucket_dir(part.get("bucket_id")))
        return {
            "type": "text",
            "text": "".join(read_bucket(bucket_dir))
        }
    if is_valid_audio(filename=filename):
        return {
            "type": "input_audio",
            "input_audio": {
                "data": render_media(**part, as_base64=True),
                "format": os.path.splitext(filename)[1][1:]
            }
        }
    if is_valid_media(filename=filename):
        return {
            "type": "image_url",
            "image_url": {"url": render_media(**part)}
        }

def merge_media(media: list, messages: list) -> Iterator:
    buffer = []
    # Read media from the last user message
    for message in messages:
        if message.get("role") == "user":
            content = message.get("content")
            if isinstance(content, list):
                for part in content:
                    if "type" not in part and "name" in part and "text" not in part:
                        path = render_media(**part, as_path=True)
                        buffer.append((path, os.path.basename(path)))
                    elif part.get("type") == "image_url":
                        path: str = urlparse(part.get("image_url")).path
                        if path.startswith("/files/"):
                            path = get_bucket_dir(path.split(path, "/")[1:])
                            if os.path.exists(path):
                                buffer.append((Path(path), os.path.basename(path)))
                            else:
                                buffer.append((part.get("image_url"), None))
                        else:
                            buffer.append((part.get("image_url"), None))
        else:
            buffer = []
    yield from buffer
    if media is not None:
        yield from media

def render_messages(messages: Messages, media: list = None) -> Iterator:
    last_is_assistant = False
    for idx, message in enumerate(messages):
        # Remove duplicate assistant messages
        if message.get("role") == "assistant":
            if last_is_assistant:
                continue
            last_is_assistant = True
        else:
            last_is_assistant = False
        # Render content parts
        if isinstance(message.get("content"), list):
            parts = [render_part(part) for part in message["content"] if part]
            yield {
                **message,
                "content": [part for part in parts if part]
            }
        else:
            # Append media to the last message
            if media is not None and idx == len(messages) - 1:
                yield {
                    **message,
                    "content": [
                        {
                            "type": "input_audio",
                            "input_audio": to_input_audio(media_data, filename)
                        }
                        if is_valid_audio(media_data, filename) else {
                            "type": "image_url",
                            "image_url": {"url": to_data_uri(media_data)}
                        }
                        for media_data, filename in media
                        if media_data and is_valid_media(media_data, filename)
                    ] + ([{"type": "text", "text": message["content"]}] if isinstance(message["content"], str) else message["content"])
                }
            else:
                yield message


"""Utilities for G4F providers and models information."""

from __future__ import annotations

import logging
from typing import Dict, List, Any, Optional, Union
from dataclasses import dataclass

# Import g4f
try:
    from .. import g4f
    from ..g4f.Provider import __providers__, __map__
    from ..g4f.providers.base_provider import ProviderModelMixin
except ImportError as e:
    raise ImportError(
        "Could not import g4f python package. "
        "Please install it with `pip install g4f`."
    ) from e

logger = logging.getLogger(__name__)

@dataclass
class ProviderInfo:
    """Information about a G4F provider."""
    name: str
    label: str
    url: Optional[str] = None
    working: bool = False
    needs_auth: bool = False
    auth_type: Optional[str] = None
    login_url: Optional[str] = None
    supports_text: bool = True
    supports_images: bool = False
    supports_vision: bool = False
    supports_audio: bool = False
    supports_video: bool = False
    supports_stream: bool = False
    default_model: Optional[str] = None
    models: List[str] = None
    image_models: List[str] = None
    vision_models: List[str] = None
    audio_models: List[str] = None
    video_models: List[str] = None

    def __post_init__(self):
        if self.models is None:
            self.models = []
        if self.image_models is None:
            self.image_models = []
        if self.vision_models is None:
            self.vision_models = []
        if self.audio_models is None:
            self.audio_models = []
        if self.video_models is None:
            self.video_models = []

@dataclass
class ModelInfo:
    """Information about a model."""
    name: str
    provider: str
    supports_vision: bool = False
    supports_audio: bool = False
    supports_video: bool = False
    supports_image_generation: bool = False
    needs_auth: bool = False
    is_default: bool = False

class G4FUtils:
    """Utility class for G4F providers and models information."""
    
    _providers_cache: Optional[List[ProviderInfo]] = None
    _models_cache: Optional[List[ModelInfo]] = None
    
    @classmethod
    def get_providers(cls, force_refresh: bool = False) -> List[ProviderInfo]:
        """Get list of all available providers with their information.
        
        Args:
            force_refresh: Force refresh the cache
            
        Returns:
            List of ProviderInfo objects
        """
        if cls._providers_cache is None or force_refresh:
            cls._providers_cache = cls._fetch_providers()
        return cls._providers_cache
    
    @classmethod
    def get_providers_by_auth(cls, needs_auth: Optional[bool] = None) -> List[ProviderInfo]:
        """Get providers filtered by authentication requirement.
        
        Args:
            needs_auth: True for auth required, False for no auth, None for all
            
        Returns:
            Filtered list of ProviderInfo objects
        """
        providers = cls.get_providers()
        if needs_auth is None:
            return providers
        return [p for p in providers if p.needs_auth == needs_auth]
    
    @classmethod
    def get_providers_by_capability(cls, 
                                  text: bool = False,
                                  images: bool = False, 
                                  vision: bool = False,
                                  audio: bool = False,
                                  video: bool = False) -> List[ProviderInfo]:
        """Get providers filtered by capabilities.
        
        Args:
            text: Support text generation
            images: Support image generation  
            vision: Support image understanding
            audio: Support audio generation
            video: Support video generation
            
        Returns:
            Filtered list of ProviderInfo objects
        """
        providers = cls.get_providers()
        filtered = []
        
        for provider in providers:
            match = True
            if text and not provider.supports_text:
                match = False
            if images and not provider.supports_images:
                match = False
            if vision and not provider.supports_vision:
                match = False
            if audio and not provider.supports_audio:
                match = False
            if video and not provider.supports_video:
                match = False
            
            if match:
                filtered.append(provider)
                
        return filtered
    
    @classmethod
    def get_working_providers(cls) -> List[ProviderInfo]:
        """Get only working providers.
        
        Returns:
            List of working ProviderInfo objects
        """
        providers = cls.get_providers()
        return [p for p in providers if p.working]
    
    @classmethod
    def get_provider_by_name(cls, name: str) -> Optional[ProviderInfo]:
        """Get provider information by name.
        
        Args:
            name: Provider name
            
        Returns:
            ProviderInfo object or None if not found
        """
        providers = cls.get_providers()
        for provider in providers:
            if provider.name.lower() == name.lower():
                return provider
        return None
    
    @classmethod
    def get_models(cls, provider_name: Optional[str] = None, force_refresh: bool = False) -> List[ModelInfo]:
        """Get list of all available models.
        
        Args:
            provider_name: Filter by specific provider name
            force_refresh: Force refresh the cache
            
        Returns:
            List of ModelInfo objects
        """
        if cls._models_cache is None or force_refresh:
            cls._models_cache = cls._fetch_models()
        
        models = cls._models_cache
        if provider_name:
            models = [m for m in models if m.provider.lower() == provider_name.lower()]
        
        return models
    
    @classmethod
    def get_models_by_capability(cls, 
                               vision: bool = False,
                               audio: bool = False,
                               video: bool = False,
                               image_generation: bool = False,
                               needs_auth: Optional[bool] = None) -> List[ModelInfo]:
        """Get models filtered by capabilities.
        
        Args:
            vision: Support vision/image understanding
            audio: Support audio processing
            video: Support video processing
            image_generation: Support image generation
            needs_auth: Filter by auth requirement
            
        Returns:
            Filtered list of ModelInfo objects
        """
        models = cls.get_models()
        filtered = []
        
        for model in models:
            match = True
            if vision and not model.supports_vision:
                match = False
            if audio and not model.supports_audio:
                match = False
            if video and not model.supports_video:
                match = False
            if image_generation and not model.supports_image_generation:
                match = False
            if needs_auth is not None and model.needs_auth != needs_auth:
                match = False
            
            if match:
                filtered.append(model)
                
        return filtered
    
    @classmethod
    def get_image_generation_models(cls) -> List[ModelInfo]:
        """Get all models that support image generation.
        
        Returns:
            List of ModelInfo objects that support image generation
        """
        return cls.get_models_by_capability(image_generation=True)
    
    @classmethod
    def get_vision_models(cls) -> List[ModelInfo]:
        """Get all models that support vision/image understanding.
        
        Returns:
            List of ModelInfo objects that support vision
        """
        return cls.get_models_by_capability(vision=True)
    
    @classmethod
    def _fetch_providers(cls) -> List[ProviderInfo]:
        """Fetch provider information from g4f."""
        providers = []
        
        try:
            for provider_class in __providers__:
                if not hasattr(provider_class, '__name__'):
                    continue
                    
                try:
                    info = cls._extract_provider_info(provider_class)
                    providers.append(info)
                except Exception as e:
                    logger.warning(f"Error extracting info for provider {getattr(provider_class, '__name__', 'Unknown')}: {e}")
                    continue
                    
        except Exception as e:
            logger.error(f"Error fetching providers: {e}")
            
        return providers
    
    @classmethod
    def _extract_provider_info(cls, provider_class) -> ProviderInfo:
        """Extract information from a provider class."""
        name = provider_class.__name__
        
        # Basic info
        label = getattr(provider_class, 'label', name)
        url = getattr(provider_class, 'url', None)
        working = getattr(provider_class, 'working', False)
        needs_auth = getattr(provider_class, 'needs_auth', False)
        login_url = getattr(provider_class, 'login_url', None)
        supports_stream = getattr(provider_class, 'supports_stream', False)
        
        # Determine auth type
        auth_type = None
        if needs_auth:
            if hasattr(provider_class, 'use_nodriver') and provider_class.use_nodriver:
                auth_type = "nodriver"
            elif login_url and "api" in login_url.lower():
                auth_type = "api_key"
            elif login_url:
                auth_type = "cookies"
            else:
                auth_type = "unknown"
        
        # Models and capabilities
        default_model = getattr(provider_class, 'default_model', None)
        
        # Get models safely
        models = []
        image_models = []
        vision_models = []
        audio_models = []
        video_models = []
        
        try:
            if hasattr(provider_class, 'get_models') and callable(provider_class.get_models):
                models = provider_class.get_models() or []
            elif hasattr(provider_class, 'models'):
                models = provider_class.models or []
        except Exception as e:
            logger.debug(f"Could not get models for {name}: {e}")
            models = []
        
        # Image models
        try:
            image_models = getattr(provider_class, 'image_models', []) or []
        except Exception:
            image_models = []
        
        # Vision models
        try:
            vision_models = getattr(provider_class, 'vision_models', []) or []
        except Exception:
            vision_models = []
        
        # Audio models
        try:
            audio_models = getattr(provider_class, 'audio_models', []) or []
        except Exception:
            audio_models = []
        
        # Video models
        try:
            video_models = getattr(provider_class, 'video_models', []) or []
        except Exception:
            video_models = []
        
        # Determine capabilities
        supports_images = len(image_models) > 0
        supports_vision = len(vision_models) > 0 or getattr(provider_class, 'default_vision_model', None) is not None
        supports_audio = len(audio_models) > 0
        supports_video = len(video_models) > 0
        
        return ProviderInfo(
            name=name,
            label=label,
            url=url,
            working=working,
            needs_auth=needs_auth,
            auth_type=auth_type,
            login_url=login_url,
            supports_text=len(models) > 0,
            supports_images=supports_images,
            supports_vision=supports_vision,
            supports_audio=supports_audio,
            supports_video=supports_video,
            supports_stream=supports_stream,
            default_model=default_model,
            models=models,
            image_models=image_models,
            vision_models=vision_models,
            audio_models=audio_models,
            video_models=video_models
        )
    
    @classmethod
    def _fetch_models(cls) -> List[ModelInfo]:
        """Fetch model information from all providers."""
        models = []
        providers = cls.get_providers()
        
        for provider in providers:
            # Add text models
            for model_name in provider.models:
                model_info = ModelInfo(
                    name=model_name,
                    provider=provider.name,
                    supports_vision=model_name in provider.vision_models,
                    supports_audio=model_name in provider.audio_models,
                    supports_video=model_name in provider.video_models,
                    supports_image_generation=model_name in provider.image_models,
                    needs_auth=provider.needs_auth,
                    is_default=model_name == provider.default_model
                )
                models.append(model_info)
            
            # Add image models (if not already included)
            for model_name in provider.image_models:
                if model_name not in provider.models:
                    model_info = ModelInfo(
                        name=model_name,
                        provider=provider.name,
                        supports_vision=False,
                        supports_audio=False,
                        supports_video=False,
                        supports_image_generation=True,
                        needs_auth=provider.needs_auth,
                        is_default=False
                    )
                    models.append(model_info)
        
        return models
    
    @classmethod
    def print_providers_summary(cls):
        """Print a summary of all providers."""
        providers = cls.get_providers()
        working_providers = cls.get_working_providers()
        auth_required = cls.get_providers_by_auth(needs_auth=True)
        no_auth = cls.get_providers_by_auth(needs_auth=False)
        image_providers = cls.get_providers_by_capability(images=True)
        vision_providers = cls.get_providers_by_capability(vision=True)
        
        print("=== G4F Providers Summary ===")
        print(f"Total providers: {len(providers)}")
        print(f"Working providers: {len(working_providers)}")
        print(f"Auth required: {len(auth_required)}")
        print(f"No auth required: {len(no_auth)}")
        print(f"Image generation: {len(image_providers)}")
        print(f"Vision support: {len(vision_providers)}")
        
        print("\n=== Working Providers by Auth Type ===")
        auth_types = {}
        for provider in working_providers:
            auth_type = "No Auth" if not provider.needs_auth else (provider.auth_type or "Unknown")
            if auth_type not in auth_types:
                auth_types[auth_type] = []
            auth_types[auth_type].append(provider.name)
        
        for auth_type, provider_names in auth_types.items():
            print(f"\n{auth_type} ({len(provider_names)}):")
            for name in sorted(provider_names):
                print(f"  - {name}")
    
    @classmethod
    def print_models_summary(cls):
        """Print a summary of all models."""
        models = cls.get_models()
        image_models = cls.get_image_generation_models()
        vision_models = cls.get_vision_models()
        
        # Group by provider
        provider_models = {}
        for model in models:
            if model.provider not in provider_models:
                provider_models[model.provider] = []
            provider_models[model.provider].append(model)
        
        print("=== G4F Models Summary ===")
        print(f"Total models: {len(models)}")
        print(f"Image generation models: {len(image_models)}")
        print(f"Vision models: {len(vision_models)}")
        print(f"Providers with models: {len(provider_models)}")
        
        print("\n=== Models by Provider ===")
        for provider_name in sorted(provider_models.keys()):
            provider_model_list = provider_models[provider_name]
            print(f"\n{provider_name} ({len(provider_model_list)} models):")
            
            # Group by capability
            text_models = [m for m in provider_model_list if not m.supports_image_generation]
            img_models = [m for m in provider_model_list if m.supports_image_generation]
            
            if text_models:
                print(f"  Text: {', '.join([m.name for m in text_models[:5]])}" + 
                      (f" ... (+{len(text_models)-5} more)" if len(text_models) > 5 else ""))
            
            if img_models:
                print(f"  Image: {', '.join([m.name for m in img_models])}")


def get_providers(**filters) -> List[ProviderInfo]:
    """Get providers with optional filters.
    
    Args:
        **filters: Keyword arguments for filtering (working, needs_auth, supports_images, etc.)
        
    Returns:
        List of filtered ProviderInfo objects
    """
    providers = G4FUtils.get_providers()
    
    for key, value in filters.items():
        if hasattr(ProviderInfo, key):
            providers = [p for p in providers if getattr(p, key) == value]
    
    return providers

def get_models(**filters) -> List[ModelInfo]:
    """Get models with optional filters.
    
    Args:
        **filters: Keyword arguments for filtering (provider, supports_vision, etc.)
        
    Returns:
        List of filtered ModelInfo objects
    """
    models = G4FUtils.get_models()
    
    for key, value in filters.items():
        if hasattr(ModelInfo, key):
            models = [m for m in models if getattr(m, key) == value]
    
    return models

def print_summary():
    """Print a summary of providers and models."""
    G4FUtils.print_providers_summary()
    print("\n" + "="*50 + "\n")
    G4FUtils.print_models_summary()



from __future__ import annotations

import os

try:
    from gtts import gTTS as gTTS_Service
    has_gtts = True
except ImportError:
    has_gtts = False

from ...typing import AsyncResult, Messages
from ...providers.response import AudioResponse
from ...image.copy_images import get_filename, get_media_dir, ensure_media_dir
from ..base_provider import AsyncGeneratorProvider, ProviderModelMixin
from ..helper import get_last_message

locals = {
    "en-AU": ["English (Australia)", "en", "com.au"],
    "en-GB": ["English (United Kingdom)", "en", "co.uk"],
    "en-US": ["English (United States)", "en", "us"],
    "en-CA": ["English (Canada)", "en", "ca"],
    "en-IN": ["English (India)", "en", "co.in"],
    "en-IE": ["English (Ireland)", "en", "ie"],
    "en-ZA": ["English (South Africa)", "en", "co.za"],
    "en-NG": ["English (Nigeria)", "en", "com.ng"],
    "fr-CA": ["French (Canada)", "fr", "ca"],
    "fr-FR": ["French (France)", "fr", "fr"],
    "de-DE": ["German (Germany)", "de", "de"],
    "zh-CN": ["Mandarin (China Mainland)", "zh-CN", "com"],
    "zh-TW": ["Mandarin (Taiwan)", "zh-TW", "com"],
    "pt-BR": ["Portuguese (Brazil)", "pt", "com.br"],
    "pt-PT": ["Portuguese (Portugal)", "pt", "pt"],
    "es-MX": ["Spanish (Mexico)", "es", "com.mx"],
    "es-ES": ["Spanish (Spain)", "es", "es"],
    "es-US": ["Spanish (United States)", "es", "us"],
}
models = {locale[0]: {"lang": locale[1], "tld": locale[2]} for locale in locals.values()}

class gTTS(AsyncGeneratorProvider, ProviderModelMixin):
    label = "gTTS (Google Text-to-Speech)"
    working = has_gtts
    model_id = "google-tts"
    default_language = "en"
    default_tld = "com"
    default_format = "mp3"

    default_model = "en-US"
    models = list(models.keys())
    audio_models = models

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        prompt: str = None,
        audio: dict = {},
        **kwargs
    ) -> AsyncResult:
        prompt = get_last_message(messages, prompt)
        if not prompt:
            raise ValueError("Prompt is empty.")
        format = audio.get("format", cls.default_format)
        filename = get_filename([cls.model_id], prompt, f".{format}", prompt)
        target_path = os.path.join(get_media_dir(), filename)
        ensure_media_dir()

        if "language" in audio:
            model = locals[audio["language"]][0] if audio["language"] in locals else model

        gTTS_Service(
            prompt,
            **{
                "lang": audio.get("language", cls.default_language),
                "tld": audio.get("tld", cls.default_tld),
                "slow": audio.get("slow", False),
                **models.get(model, {})
            }
        ).save(target_path)

        yield AudioResponse(f"/media/{filename}", audio=audio, text=prompt)


"""Image generation functionality for G4F integration."""

from __future__ import annotations

import logging
import asyncio
import base64
from io import BytesIO
from typing import Any, Dict, List, Optional, Union, Literal
from pathlib import Path

from langchain_core.utils import get_pydantic_field_names
from langchain_core.utils.utils import _build_model_kwargs
from pydantic import BaseModel, Field, ConfigDict, model_validator
from typing_extensions import Self

# Import g4f
try:
    from .. import g4f
    from ..g4f.client import Client, AsyncClient
    from ..g4f.Provider import BaseProvider
except ImportError as e:
    raise ImportError(
        "Could not import g4f python package. "
        "Please install it with `pip install g4f`."
    ) from e

# Optional PIL import for image handling
try:
    from PIL import Image
    HAS_PIL = True
except ImportError:
    HAS_PIL = False

logger = logging.getLogger(__name__)

class ImageGenerationResponse:
    """Response object for image generation."""
    
    def __init__(self, 
                 url: Optional[str] = None, 
                 data: Optional[bytes] = None,
                 alt: Optional[str] = None,
                 prompt: Optional[str] = None,
                 model: Optional[str] = None,
                 provider: Optional[str] = None):
        self.url = url
        self.data = data
        self.alt = alt
        self.prompt = prompt
        self.model = model
        self.provider = provider
    
    def save(self, path: Union[str, Path]) -> None:
        """Save image to file."""
        path = Path(path)
        
        if self.data:
            # Save binary data directly
            with open(path, 'wb') as f:
                f.write(self.data)
        elif self.url:
            # Download from URL
            import requests
            response = requests.get(self.url)
            response.raise_for_status()
            with open(path, 'wb') as f:
                f.write(response.content)
        else:
            raise ValueError("No image data or URL available")
    
    def to_pil(self) -> 'Image.Image':
        """Convert to PIL Image object."""
        if not HAS_PIL:
            raise ImportError("PIL is required for to_pil(). Install with: pip install Pillow")
        
        if self.data:
            return Image.open(BytesIO(self.data))
        elif self.url:
            import requests
            response = requests.get(self.url)
            response.raise_for_status()
            return Image.open(BytesIO(response.content))
        else:
            raise ValueError("No image data or URL available")
    
    def to_base64(self) -> str:
        """Convert image to base64 string."""
        if self.data:
            return base64.b64encode(self.data).decode('utf-8')
        elif self.url:
            import requests
            response = requests.get(self.url)
            response.raise_for_status()
            return base64.b64encode(response.content).decode('utf-8')
        else:
            raise ValueError("No image data or URL available")
    
    def show(self) -> None:
        """Display image using PIL."""
        if not HAS_PIL:
            raise ImportError("PIL is required for show(). Install with: pip install Pillow")
        
        img = self.to_pil()
        img.show()
    
    def __repr__(self) -> str:
        return f"ImageGenerationResponse(url={self.url}, alt='{self.alt}', model='{self.model}', provider='{self.provider}')"

class ImageG4F(BaseModel):
    """G4F Image generation integration.

    This class provides image generation capabilities using the G4F library.

    Setup:
        Install ``g4f`` and optionally set environment variables.

        .. code-block:: bash

            pip install -U g4f

    Key init args:
        model: str
            Name of the image model to use (e.g., "flux", "dall-e-3").
        provider: Any
            G4F provider to use for image generation.
        api_key: Optional[str]
            API key for providers that require authentication.
        response_format: str
            Format for response ("url" or "b64_json").

    Example:
        .. code-block:: python

            from langchain_g4f import ImageG4F
            import g4f

            image_gen = ImageG4F(
                model="flux",
                provider=g4f.Provider.HuggingFaceMedia,
                api_key="your-api-key",
            )

            response = image_gen.generate("a white siamese cat")
            response.save("cat.png")
            response.show()  # Display using PIL
    """

    model_name: str = Field(default="flux", alias="model")
    """Image model name to use."""
    
    provider: Any = Field(default=None)
    """G4F provider to use. If None, g4f will auto-select."""
    
    api_key: Optional[str] = Field(default=None)
    """API key for providers that require authentication."""
    
    response_format: Literal["url", "b64_json"] = "url"
    """Response format for generated images."""
    
    size: Optional[str] = Field(default=None)
    """Image size (e.g., "1024x1024")."""
    
    quality: Optional[str] = Field(default=None)
    """Image quality setting."""
    
    style: Optional[str] = Field(default=None)
    """Image style setting."""
    
    n: int = 1
    """Number of images to generate."""
    
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Additional parameters for the g4f create call."""

    model_config = ConfigDict(
        populate_by_name=True,
        arbitrary_types_allowed=True,
        exclude={"provider"},
    )

    @model_validator(mode="before")
    @classmethod
    def build_extra(cls, values: Dict[str, Any]) -> Any:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        values = _build_model_kwargs(values, all_required_field_names)
        return values

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that the g4f library is available."""
        # Validate provider if specified
        if self.provider is not None:
            if not hasattr(self.provider, 'create_completion') and not hasattr(self.provider, 'create_async_generator'):
                logger.warning(f"Provider {self.provider} may not be a valid g4f provider")
        
        return self

    def _prepare_params(self, **kwargs: Any) -> Dict[str, Any]:
        """Prepare parameters for g4f image generation call."""
        params = {
            "model": self.model_name,
            "response_format": self.response_format,
            "n": self.n,
            **self.model_kwargs,
            **kwargs,
        }
        
        if self.size:
            params["size"] = self.size
        if self.quality:
            params["quality"] = self.quality
        if self.style:
            params["style"] = self.style
        if self.provider is not None:
            params["provider"] = self.provider
        if self.api_key is not None:
            params["api_key"] = self.api_key
            
        return params

    def generate(self, 
                 prompt: str, 
                 **kwargs: Any) -> Union[ImageGenerationResponse, List[ImageGenerationResponse]]:
        """Generate image(s) from text prompt.
        
        Args:
            prompt: Text description of the image to generate
            **kwargs: Additional parameters
            
        Returns:
            ImageGenerationResponse or list of responses
        """
        params = self._prepare_params(**kwargs)
        
        try:
            client = Client()
            response = client.images.generate(
                prompt=prompt,
                **params
            )
            
            return self._process_response(response, prompt)
            
        except Exception as e:
            logger.error(f"Error in g4f image generation: {e}")
            raise

    async def agenerate(self, 
                       prompt: str, 
                       **kwargs: Any) -> Union[ImageGenerationResponse, List[ImageGenerationResponse]]:
        """Async generate image(s) from text prompt.
        
        Args:
            prompt: Text description of the image to generate
            **kwargs: Additional parameters
            
        Returns:
            ImageGenerationResponse or list of responses
        """
        params = self._prepare_params(**kwargs)
        
        try:
            client = AsyncClient()
            response = await client.images.generate(
                prompt=prompt,
                **params
            )
            
            return self._process_response(response, prompt)
            
        except Exception as e:
            logger.error(f"Error in g4f async image generation: {e}")
            raise

    def create_variation(self, 
                        image: Union[str, Path, bytes], 
                        prompt: Optional[str] = None,
                        **kwargs: Any) -> Union[ImageGenerationResponse, List[ImageGenerationResponse]]:
        """Create variations of an existing image.
        
        Args:
            image: Source image (file path, URL, or bytes)
            prompt: Optional prompt for variation
            **kwargs: Additional parameters
            
        Returns:
            ImageGenerationResponse or list of responses
        """
        params = self._prepare_params(**kwargs)
        
        try:
            client = Client()
            response = client.images.create_variation(
                image=image,
                prompt=prompt,
                **params
            )
            
            return self._process_response(response, prompt or "Image variation")
            
        except Exception as e:
            logger.error(f"Error in g4f image variation: {e}")
            raise

    async def acreate_variation(self, 
                               image: Union[str, Path, bytes], 
                               prompt: Optional[str] = None,
                               **kwargs: Any) -> Union[ImageGenerationResponse, List[ImageGenerationResponse]]:
        """Async create variations of an existing image.
        
        Args:
            image: Source image (file path, URL, or bytes)
            prompt: Optional prompt for variation
            **kwargs: Additional parameters
            
        Returns:
            ImageGenerationResponse or list of responses
        """
        params = self._prepare_params(**kwargs)
        
        try:
            client = AsyncClient()
            response = await client.images.create_variation(
                image=image,
                prompt=prompt,
                **params
            )
            
            return self._process_response(response, prompt or "Image variation")
            
        except Exception as e:
            logger.error(f"Error in g4f async image variation: {e}")
            raise

    def _process_response(self, response: Any, prompt: str) -> Union[ImageGenerationResponse, List[ImageGenerationResponse]]:
        """Process the response from g4f image generation."""
        results = []
        
        try:
            # Handle different response formats
            if hasattr(response, 'data') and response.data:
                # Standard OpenAI-like response
                for item in response.data:
                    if hasattr(item, 'url') and item.url:
                        result = ImageGenerationResponse(
                            url=item.url,
                            alt=prompt,
                            prompt=prompt,
                            model=self.model_name,
                            provider=str(self.provider) if self.provider else None
                        )
                    elif hasattr(item, 'b64_json') and item.b64_json:
                        # Base64 encoded image
                        data = base64.b64decode(item.b64_json)
                        result = ImageGenerationResponse(
                            data=data,
                            alt=prompt,
                            prompt=prompt,
                            model=self.model_name,
                            provider=str(self.provider) if self.provider else None
                        )
                    else:
                        # Fallback - convert item to string (might be URL)
                        result = ImageGenerationResponse(
                            url=str(item),
                            alt=prompt,
                            prompt=prompt,
                            model=self.model_name,
                            provider=str(self.provider) if self.provider else None
                        )
                    results.append(result)
            
            elif hasattr(response, 'url'):
                # Direct URL response
                result = ImageGenerationResponse(
                    url=response.url,
                    alt=prompt,
                    prompt=prompt,
                    model=self.model_name,
                    provider=str(self.provider) if self.provider else None
                )
                results.append(result)
            
            elif isinstance(response, str):
                # Simple string URL response
                result = ImageGenerationResponse(
                    url=response,
                    alt=prompt,
                    prompt=prompt,
                    model=self.model_name,
                    provider=str(self.provider) if self.provider else None
                )
                results.append(result)
            
            else:
                # Fallback - try to convert to string
                result = ImageGenerationResponse(
                    url=str(response),
                    alt=prompt,
                    prompt=prompt,
                    model=self.model_name,
                    provider=str(self.provider) if self.provider else None
                )
                results.append(result)
        
        except Exception as e:
            logger.error(f"Error processing image response: {e}")
            # Create a fallback response
            result = ImageGenerationResponse(
                url=str(response) if response else None,
                alt=prompt,
                prompt=prompt,
                model=self.model_name,
                provider=str(self.provider) if self.provider else None
            )
            results.append(result)
        
        # Return single result or list based on number generated
        if self.n == 1:
            return results[0] if results else ImageGenerationResponse(prompt=prompt, model=self.model_name)
        else:
            return results

    @classmethod
    def get_available_models(cls) -> List[str]:
        """Get list of available image generation models.
        
        Returns:
            List of model names that support image generation
        """
        try:
            from ..core.utils import G4FUtils
            models = G4FUtils.get_image_generation_models()
            return [model.name for model in models]
        except ImportError:
            # Fallback list of common image models
            return [
                "flux", "flux-pro", "flux-dev", "flux-schnell",
                "dall-e-3", "gpt-image", 
                "sdxl-1.0", "sdxl-turbo", "sd-3.5-large",
                "midjourney"
            ]

    @classmethod
    def get_available_providers(cls) -> List[str]:
        """Get list of providers that support image generation.
        
        Returns:
            List of provider names that support image generation
        """
        try:
            from ..core.utils import G4FUtils
            providers = G4FUtils.get_providers_by_capability(images=True)
            return [provider.name for provider in providers]
        except ImportError:
            # Fallback list of common image providers
            return [
                "HuggingFaceMedia", "PollinationsAI", "OpenaiChat",
                "Copilot", "Gemini", "You", "MicrosoftDesigner"
            ]

# Convenience functions
def generate_image(prompt: str, 
                  model: str = "flux",
                  provider: Optional[Any] = None,
                  **kwargs) -> ImageGenerationResponse:
    """Generate a single image from text prompt.
    
    Args:
        prompt: Text description of the image
        model: Image model to use
        provider: G4F provider to use
        **kwargs: Additional parameters
        
    Returns:
        ImageGenerationResponse object
    """
    image_gen = ImageG4F(model=model, provider=provider, **kwargs)
    return image_gen.generate(prompt)

async def agenerate_image(prompt: str, 
                         model: str = "flux",
                         provider: Optional[Any] = None,
                         **kwargs) -> ImageGenerationResponse:
    """Async generate a single image from text prompt.
    
    Args:
        prompt: Text description of the image
        model: Image model to use
        provider: G4F provider to use
        **kwargs: Additional parameters
        
    Returns:
        ImageGenerationResponse object
    """
    image_gen = ImageG4F(model=model, provider=provider, **kwargs)
    return await image_gen.agenerate(prompt)

def batch_generate_images(prompts: List[str], 
                         model: str = "flux",
                         provider: Optional[Any] = None,
                         **kwargs) -> List[ImageGenerationResponse]:
    """Generate multiple images from text prompts.
    
    Args:
        prompts: List of text descriptions
        model: Image model to use
        provider: G4F provider to use
        **kwargs: Additional parameters
        
    Returns:
        List of ImageGenerationResponse objects
    """
    image_gen = ImageG4F(model=model, provider=provider, **kwargs)
    results = []
    
    for prompt in prompts:
        try:
            result = image_gen.generate(prompt)
            results.append(result)
        except Exception as e:
            logger.error(f"Error generating image for prompt '{prompt}': {e}")
            # Add a failed response
            results.append(ImageGenerationResponse(prompt=prompt, model=model))
    
    return results

async def abatch_generate_images(prompts: List[str], 
                                model: str = "flux",
                                provider: Optional[Any] = None,
                                **kwargs) -> List[ImageGenerationResponse]:
    """Async generate multiple images from text prompts.
    
    Args:
        prompts: List of text descriptions
        model: Image model to use
        provider: G4F provider to use
        **kwargs: Additional parameters
        
    Returns:
        List of ImageGenerationResponse objects
    """
    image_gen = ImageG4F(model=model, provider=provider, **kwargs)
    
    async def generate_single(prompt: str) -> ImageGenerationResponse:
        try:
            return await image_gen.agenerate(prompt)
        except Exception as e:
            logger.error(f"Error generating image for prompt '{prompt}': {e}")
            return ImageGenerationResponse(prompt=prompt, model=model)
    
    # Generate all images concurrently
    tasks = [generate_single(prompt) for prompt in prompts]
    return await asyncio.gather(*tasks)


class G4FImageGenerator:
    """High-level interface for G4F image generation."""
    
    def __init__(self, model: str = "flux", provider: Optional[Any] = None, **kwargs):
        """Initialize image generator.
        
        Args:
            model: Image model to use
            provider: G4F provider to use
            **kwargs: Additional parameters
        """
        self.model = model
        self.provider = provider
        self.kwargs = kwargs
        self._generator = ImageG4F(model=model, provider=provider, **kwargs)
    
    def generate(self, prompt: str) -> ImageGenerationResponse:
        """Generate image from text prompt.
        
        Args:
            prompt: Text description of the image to generate
            
        Returns:
            ImageGenerationResponse object
        """
        return self._generator.generate(prompt)
    
    async def agenerate(self, prompt: str) -> ImageGenerationResponse:
        """Async generate image from text prompt.
        
        Args:
            prompt: Text description of the image to generate
            
        Returns:
            ImageGenerationResponse object
        """
        return await self._generator.agenerate(prompt)
    
    def batch_generate(self, prompts: List[str]) -> List[ImageGenerationResponse]:
        """Generate multiple images from text prompts.
        
        Args:
            prompts: List of text descriptions
            
        Returns:
            List of ImageGenerationResponse objects
        """
        return batch_generate_images(prompts, self.model, self.provider, **self.kwargs)
    
    async def abatch_generate(self, prompts: List[str]) -> List[ImageGenerationResponse]:
        """Async generate multiple images from text prompts.
        
        Args:
            prompts: List of text descriptions
            
        Returns:
            List of ImageGenerationResponse objects
        """
        return await abatch_generate_images(prompts, self.model, self.provider, **self.kwargs)



from __future__ import annotations

from ..models import ModelUtils, ImageModel, VisionModel
from ..Provider import ProviderUtils
from ..providers.types import ProviderType

class ClientModels():
    def __init__(self, client, provider: ProviderType = None, media_provider: ProviderType = None):
        self.client = client
        self.provider = provider
        self.media_provider = media_provider

    def get(self, name, default=None) -> ProviderType:
        if name in ModelUtils.convert:
            return ModelUtils.convert[name].best_provider
        if name in ProviderUtils.convert:
            return ProviderUtils.convert[name]
        return default

    def get_all(self, api_key: str = None, **kwargs) -> list[str]:
        if self.provider is None:
            return []
        if api_key is None:
            api_key = self.client.api_key
        return self.provider.get_models(
            **kwargs,
            **{} if api_key is None else {"api_key": api_key}
        )

    def get_vision(self, **kwargs) -> list[str]:
        if self.provider is None:
            return [model_id for model_id, model in ModelUtils.convert.items() if isinstance(model, VisionModel)]
        self.get_all(**kwargs)
        if hasattr(self.provider, "vision_models"):
            return self.provider.vision_models
        return []

    def get_media(self, api_key: str = None, **kwargs) -> list[str]:
        if self.media_provider is None:
            return []
        if api_key is None:
            api_key = self.client.api_key
        return self.media_provider.get_models(
            **kwargs,
            **{} if api_key is None else {"api_key": api_key}
        )

    def get_image(self, **kwargs) -> list[str]:
        if self.media_provider is None:
            return [model_id for model_id, model in ModelUtils.convert.items() if isinstance(model, ImageModel)]
        self.get_media(**kwargs)
        if hasattr(self.media_provider, "image_models"):
            return self.media_provider.image_models
        return []

    def get_video(self, **kwargs) -> list[str]:
        if self.media_provider is None:
            return []
        self.get_media(**kwargs)
        if hasattr(self.media_provider, "video_models"):
            return self.media_provider.video_models
        return []